import datetime
import enum
import math
import os.path
import pathlib
import re
from dataclasses import dataclass
from decimal import Decimal
from fractions import Fraction

import yuio.json_schema
import yuio.parse
import yuio.secret
import yuio.widget

import jsonschema
import pytest

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t


class TestStrParsingContext:
    @pytest.mark.parametrize(
        ("s", "delim", "expected", "maxsplit"),
        [
            ("", None, [], -1),
            ("", ",", [("", 0, 0)], -1),
            ("asd", None, [("asd", 0, 3)], -1),
            ("asd dsa", None, [("asd", 0, 3), ("dsa", 4, 7)], -1),
            ("  asd  dsa  ", None, [("asd", 2, 5), ("dsa", 7, 10)], -1),
            ("asd  dsa  ", None, [("asd", 0, 3), ("dsa", 5, 8)], -1),
            ("  asd  dsa", None, [("asd", 2, 5), ("dsa", 7, 10)], -1),
            ("asd", None, [("asd", 0, 3)], 0),
            ("asd  ", None, [("asd  ", 0, 5)], 0),
            ("  asd", None, [("asd", 2, 5)], 0),
            ("  asd  ", None, [("asd  ", 2, 7)], 0),
            ("  asd", None, [("asd", 2, 5)], 1),
            ("asd  ", None, [("asd", 0, 3)], 1),
            ("  asd  ", None, [("asd", 2, 5)], 1),
            ("  asd  dsa   ", None, [("asd", 2, 5), ("dsa   ", 7, 13)], 1),
            ("  asd  dsa", None, [("asd", 2, 5), ("dsa", 7, 10)], 2),
            ("a,b,c", ",", [("a,b,c", 0, 5)], 0),
            ("a,b,c", ",", [("a", 0, 1), ("b,c", 2, 5)], 1),
            ("a,b,c", ",", [("a", 0, 1), ("b", 2, 3), ("c", 4, 5)], -1),
            (
                ",a,,b,",
                ",",
                [("", 0, 0), ("a", 1, 2), ("", 3, 3), ("b", 4, 5), ("", 6, 6)],
                -1,
            ),
        ],
    )
    def test_split(self, s, delim, expected, maxsplit):
        ctx = yuio.parse.StrParsingContext(s)

        results = [
            (ctx.value, ctx.start, ctx.end)
            for ctx in ctx.split(delim, maxsplit=maxsplit)
        ]
        assert results == expected

        raw_results = [res[0] for res in results]
        assert raw_results == s.split(delim, maxsplit=maxsplit), (
            "should match built-in str.split"
        )

    @pytest.mark.parametrize(
        ("s", "chars", "expected"),
        [
            ("", None, ("", 0, 0)),
            ("asd", None, ("asd", 0, 3)),
            (" asd ", None, ("asd", 1, 4)),
            ("\n\tasd  ", None, ("asd", 2, 5)),
            ("   ", None, ("", 3, 3)),
            ("", ".,", ("", 0, 0)),
            ("asd", ".,", ("asd", 0, 3)),
            (".,.asd.,.", ".,", ("asd", 3, 6)),
            (".,.", ".,", ("", 3, 3)),
        ],
    )
    def test_strip(self, s, chars, expected):
        ctx = yuio.parse.StrParsingContext(s)
        strip = ctx.strip(chars)
        result = (strip.value, strip.start, strip.end)
        assert result == expected
        assert strip.value == ctx.value.strip(chars)

    @pytest.mark.parametrize(
        ("s", "expected"),
        [
            ("", ("", 0, 0)),
            ("asd", ("asd", 0, 3)),
            (" asd ", ("asd", 1, 4)),
            ("\n\tasd  ", ("asd", 2, 5)),
            ("   ", ("   ", 0, 3)),
        ],
    )
    def test_strip_if_not_spaces(self, s, expected):
        ctx = yuio.parse.StrParsingContext(s)
        strip = ctx.strip_if_non_space()
        result = (strip.value, strip.start, strip.end)
        assert result == expected


class TestConfigParsingContext:
    def test_descend(self):
        ctx = yuio.parse.ConfigParsingContext(1)

        assert ctx.value == 1
        assert ctx.key is None
        assert ctx.desc is None
        assert ctx.parent is None

        ctx2 = ctx.descend(2, "k2", "d2")

        assert ctx2.value == 2
        assert ctx2.key == "k2"
        assert ctx2.desc == "d2"
        assert ctx2.parent is ctx

    def test_make_path(self):
        ctx = yuio.parse.ConfigParsingContext(1)

        assert ctx.make_path() == []
        ctx2 = ctx.descend(2, "k2", "d2")
        assert ctx2.make_path() == [("k2", "d2")]
        ctx3 = ctx2.descend(3, "k3")
        assert ctx3.make_path() == [("k2", "d2"), ("k3", None)]
        ctx4 = ctx3.descend(3, "k4")
        assert ctx4.make_path() == [("k2", "d2"), ("k3", None), ("k4", None)]


@pytest.mark.parametrize(
    ("s", "pos", "expected_s", "expected_pos"),
    [
        ("", (0, 0), '""', (1, 1)),
        ("asd", (0, 0), '"asd"', (1, 1)),
        ("asd", (0, 3), '"asd"', (1, 4)),
        ("asd", (3, 3), '"asd"', (4, 4)),
        ("asd\ndsa", (0, 3), '"asd\\ndsa"', (1, 4)),
        ("asd\ndsa", (0, 4), '"asd\\ndsa"', (1, 6)),
        ("asd\ndsa", (4, 6), '"asd\\ndsa"', (6, 8)),
    ],
)
def test_repr_and_adjust_pos(s, pos, expected_s, expected_pos):
    s2, pos2 = yuio.parse._repr_and_adjust_pos(s, pos)
    assert s2 == expected_s
    assert pos2 == expected_pos


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        ([], ""),
        ([("foo", None)], "$.foo"),
        ([("_foo", None)], "$._foo"),
        ([("-foo", None)], "$['-foo']"),
        ([("foo", None), ("bar", None)], "$.foo.bar"),
        ([("foo", None), (10, None)], "$.foo[10]"),
        ([("f\no", None), (10, None)], "$['f\\no'][10]"),
        ([("foo", "desc")], "desc"),
        ([("foo", "desc %(key)s")], "desc foo"),
        ([("foo", "desc %(key)r")], "desc 'foo'"),
        ([("x", None), ("foo", "desc %(key)r")], "$.x, desc 'foo'"),
        ([(10, None), ("foo", "desc %(key)r")], "$[10], desc 'foo'"),
        ([("foo", "desc %(key)r"), ("x", None)], "$.<desc 'foo'>.x"),
    ],
)
def test_path_renderer(path, expected, ctx):
    assert str(ctx.str(yuio.parse._PathRenderer(path))) == expected


@pytest.mark.parametrize(("as_cli"), [True, False])
@pytest.mark.parametrize(
    ("code", "pos", "a", "b"),
    [
        (
            "code",
            (0, 4),
            "> code",
            "  ~~~~",
        ),
        (
            "code",
            (2, 3),
            "> code",
            "    ~",
        ),
        (
            "some very long code very long indeed",
            (0, 36),
            "> some very long ...",
            "  ~~~~~~~~~~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (15, 36),
            "> ...code very lo...",
            "     ~~~~~~~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (3, 36),
            "> some very long ...",
            "     ~~~~~~~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (0, 9),
            "> some very long ...",
            "  ~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (3, 12),
            "> some very long ...",
            "     ~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (5, 17),
            "> ...very long co...",
            "     ~~~~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (25, 36),
            "> ...ery long indeed",
            "         ~~~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (5, 14),
            "> some very long ...",
            "       ~~~~~~~~~",
        ),
        (
            "some very long code very long indeed",
            (25, 33),
            "> ...ery long indeed",
            "         ~~~~~~~~",
        ),
        (
            "some very long code",
            (5, 9),
            "> some very long ...",
            "       ~~~~",
        ),
    ],
)
def test_code_renderer(code, pos, a, b, as_cli, ctx):
    if as_cli:
        a = "$ " + a[2:]
    col = ctx.str(yuio.parse._CodeRenderer(code, pos, as_cli=as_cli))
    assert str(col) == a + "\n" + b


class TestErrorMessages:
    def test_code(self):
        parser = yuio.parse.List(yuio.parse.Int())
        with pytest.raises(
            yuio.parse.ParsingError,
            match=re.escape("> \"10 xx 30\"\n      ~~\nCan't parse 'xx' as int"),
        ):
            parser.parse("10 xx 30")

    def test_path(self):
        parser = yuio.parse.List(yuio.parse.Int())
        with pytest.raises(
            yuio.parse.ParsingError,
            match=re.escape("In $[1]:\n  Expected int, got str: '20'"),
        ):
            parser.parse_config([10, "20", 30])


class TestStr:
    def test_basics(self):
        parser = yuio.parse.Str()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<str>"
        assert parser.describe_many() == "<str>"
        assert parser.describe_value("foo") == "foo"

    def test_json_schema(self):
        parser = yuio.parse.Str()
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.String()
        )
        assert parser.to_json_value("asd") == "asd"

    def test_parse(self):
        parser = yuio.parse.Str()
        assert parser.parse("Test") == "Test"
        assert parser.parse(" Test ") == " Test "
        assert parser.parse_config("Test") == "Test"
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_lower(self):
        parser = yuio.parse.Lower(yuio.parse.Str())
        assert parser.parse("Test") == "test"
        assert parser.parse("Test") == "test"
        assert parser.parse("ῼ") == "ῳ"
        assert parser.parse_config("Test") == "test"
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_casefold(self):
        parser = yuio.parse.CaseFold(yuio.parse.Str())
        assert parser.parse("Test") == "test"
        assert parser.parse("Test") == "test"
        assert parser.parse("ῼ") == "ωι"
        assert parser.parse_config("Test") == "test"
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_upper(self):
        parser = yuio.parse.Upper(yuio.parse.Str())
        assert parser.parse("Test") == "TEST"
        assert parser.parse("Test") == "TEST"
        assert parser.parse_config("Test") == "TEST"
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_strip(self):
        parser = yuio.parse.Strip(yuio.parse.Str())
        assert parser.parse("Test  ") == "Test"
        assert parser.parse("  Test") == "Test"
        assert parser.parse_config("  Test  ") == "Test"
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_regex(self):
        parser = yuio.parse.Regex(yuio.parse.Str(), r"^a|b$")
        assert parser.parse("a") == "a"
        assert parser.parse("b") == "b"
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Value doesn't match regex \^a\|b\$"
        ):
            parser.parse("foo")

    def test_regex_compiled(self):
        parser = yuio.parse.Regex(yuio.parse.Str(), re.compile(r"^a|b$"))
        assert parser.parse("a") == "a"
        assert parser.parse("b") == "b"
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Value doesn't match regex \^a\|b\$"
        ):
            parser.parse("foo")

    def test_many(self):
        parser = yuio.parse.Str()
        with pytest.raises(RuntimeError, match=r"unable to parse multiple values"):
            parser.parse_many([])

    def test_from_type_hint(self):
        parser = yuio.parse.from_type_hint(str)
        assert parser.parse("Test") == "Test"

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[str, yuio.parse.Str()])
        assert parser.parse("  Test  ") == "  Test  "

    def test_from_type_hint_annotated_strip(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[str, yuio.parse.Str(), yuio.parse.Strip()]
        )
        assert parser.parse("  Test  ") == "Test"

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Str conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Str()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Str will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[str, yuio.parse.Str(), yuio.parse.Str()]
            )


class TestInt:
    def test_basics(self):
        parser = yuio.parse.Int()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<int>"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(10) == "10"

    def test_json_schema(self):
        parser = yuio.parse.Int()
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Integer()
        )
        assert parser.to_json_value(10) == 10

    def test_parse(self):
        parser = yuio.parse.Int()
        assert parser.parse("1") == 1
        assert parser.parse("0x10") == 16
        assert parser.parse("0o10") == 8
        assert parser.parse("0b10") == 2
        assert parser.parse("-1") == -1
        assert parser.parse("-0x10") == -16
        assert parser.parse("-0o10") == -8
        assert parser.parse("-0b10") == -2
        assert parser.parse("  -  1  ") == -1
        assert parser.parse("  -  0x10  ") == -16
        assert parser.parse("  -  0o10  ") == -8
        assert parser.parse("  -  0b10  ") == -2
        assert parser.parse_config(1) == 1
        assert parser.parse_config(1.0) == 1
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("x")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("0x-10")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("0x 10")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("--10")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected int"):
            parser.parse_config(1.5)
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected int"):
            parser.parse_config("x")

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(int), yuio.parse.Int)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Int()])
        assert parser.parse("1") == 1

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating str with Int conflicts with default "
                "parser for this type, which is Str."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[str, yuio.parse.Int()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Int will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Int(), yuio.parse.Int()]
            )


class TestFloat:
    def test_basics(self):
        parser = yuio.parse.Float()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<float>"
        assert parser.describe_many() == "<float>"
        assert parser.describe_value(10.5) == "10.5"

    def test_json_schema(self):
        parser = yuio.parse.Float()
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Number()
        )
        assert parser.to_json_value(10.5) == 10.5

    def test_parse(self):
        parser = yuio.parse.Float()
        assert parser.parse("1.5") == 1.5
        assert parser.parse("-10") == -10.0
        assert parser.parse("2e9") == 2e9
        assert parser.parse_config(1.0) == 1.0
        assert parser.parse_config(1.5) == 1.5
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("x")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected float"):
            parser.parse_config("x")

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(float), yuio.parse.Float)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[float, yuio.parse.Float()])
        assert parser.parse("1.5") == 1.5

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Float conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Float()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Float will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[float, yuio.parse.Float(), yuio.parse.Float()]
            )


class TestBool:
    def test_basics(self):
        parser = yuio.parse.Bool()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "{yes|no}"
        assert parser.describe_or_def() == "{yes|no}"
        assert parser.describe_many() == "{yes|no}"
        assert parser.describe_value(True) == "yes"
        assert parser.describe_value(False) == "no"

    def test_json_schema(self):
        parser = yuio.parse.Bool()
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Boolean()
        )
        assert parser.to_json_value(True)

    def test_parse(self):
        parser = yuio.parse.Bool()
        assert parser.parse("y") is True
        assert parser.parse("yes") is True
        assert parser.parse("yEs") is True
        assert parser.parse("n") is False
        assert parser.parse("no") is False
        assert parser.parse("nO") is False
        with pytest.raises(yuio.parse.ParsingError, match=r"yes, no, true, or false"):
            parser.parse("Meh")
        assert parser.parse_config(True) is True
        assert parser.parse_config(False) is False
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected bool"):
            parser.parse_config("x")

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(bool), yuio.parse.Bool)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[bool, yuio.parse.Bool()])
        assert parser.parse("yes") is True

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Bool conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Bool()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Bool will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[bool, yuio.parse.Bool(), yuio.parse.Bool()]
            )


class TestEnum:
    class Cuteness(enum.Enum):
        #: Cats!
        CATS = "Cats"
        #: Dogs!
        DOGS = "Dogs"
        #: Sharkie!
        BLAHAJ = ":3"

    class Colors(enum.IntEnum):
        RED = 31
        GREEN = 32
        BLUE = 34

    def test_basics_by_value(self):
        parser = yuio.parse.Enum(self.Cuteness)
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() == [
            yuio.widget.Option(
                value=self.Cuteness.CATS,
                display_text="Cats",
                display_text_prefix="",
                display_text_suffix="",
                comment="Cats!",
                color_tag="none",
            ),
            yuio.widget.Option(
                value=self.Cuteness.DOGS,
                display_text="Dogs",
                display_text_prefix="",
                display_text_suffix="",
                comment="Dogs!",
                color_tag="none",
            ),
            yuio.widget.Option(
                value=self.Cuteness.BLAHAJ,
                display_text=":3",
                display_text_prefix="",
                display_text_suffix="",
                comment="Sharkie!",
                color_tag="none",
            ),
        ]
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "{Cats|Dogs|:3}"
        assert parser.describe_or_def() == "{Cats|Dogs|:3}"
        assert parser.describe_many() == "{Cats|Dogs|:3}"
        assert parser.describe_value(self.Cuteness.BLAHAJ) == ":3"

    def test_basics_by_name(self):
        parser = yuio.parse.Enum(self.Cuteness, by_name=True)
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "{CATS|DOGS|BLAHAJ}"
        assert parser.describe_or_def() == "{CATS|DOGS|BLAHAJ}"
        assert parser.describe_many() == "{CATS|DOGS|BLAHAJ}"
        assert parser.describe_value(self.Cuteness.BLAHAJ) == "BLAHAJ"

    def test_json_schema_by_value(self):
        parser = yuio.parse.Enum(self.Cuteness)
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref(
            "#/$defs/test.test_parse.TestEnum.Cuteness",
            "test.test_parse.TestEnum.Cuteness",
        )
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate("Cats")
        validator.validate("Dogs")
        validator.validate(":3")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("DOGS")
        assert parser.to_json_value(self.Cuteness.BLAHAJ) == ":3"

    def test_json_schema_by_name(self):
        parser = yuio.parse.Enum(self.Cuteness, by_name=True)
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref(
            "#/$defs/test.test_parse.TestEnum.Cuteness",
            "test.test_parse.TestEnum.Cuteness",
        )
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate("CATS")
        validator.validate("DOGS")
        validator.validate("BLAHAJ")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("Dogs")
        assert parser.to_json_value(self.Cuteness.BLAHAJ) == "BLAHAJ"

    def test_json_schema_name_clash(self):
        ctx = yuio.json_schema.JsonSchemaContext()
        parser = yuio.parse.Enum(self.Cuteness)
        assert parser.to_json_schema(ctx) == yuio.json_schema.Ref(
            "#/$defs/test.test_parse.TestEnum.Cuteness",
            "test.test_parse.TestEnum.Cuteness",
        )

        parser = yuio.parse.Enum(self.Cuteness, by_name=True)
        assert parser.to_json_schema(ctx) == yuio.json_schema.Ref(
            "#/$defs/test.test_parse.TestEnum.Cuteness2",
            "test.test_parse.TestEnum.Cuteness2",
        )

        parser = yuio.parse.Enum(self.Cuteness, to_dash_case=True)
        assert parser.to_json_schema(ctx) == yuio.json_schema.Ref(
            "#/$defs/test.test_parse.TestEnum.Cuteness3",
            "test.test_parse.TestEnum.Cuteness3",
        )

        parser = yuio.parse.Enum(self.Cuteness)
        assert parser.to_json_schema(ctx) == yuio.json_schema.Ref(
            "#/$defs/test.test_parse.TestEnum.Cuteness",
            "test.test_parse.TestEnum.Cuteness",
        )

    def test_json_schema_inline(self):
        parser = yuio.parse.Enum(self.Cuteness, doc_inline=True)
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.Enum(
            ["Cats", "Dogs", ":3"], ["Cats!", "Dogs!", "Sharkie!"]
        )

    def test_by_value(self):
        parser = yuio.parse.Enum(self.Cuteness)
        assert parser.parse("CATS") is self.Cuteness.CATS
        assert parser.parse("CATS") is self.Cuteness.CATS
        assert parser.parse_config("Cats") is self.Cuteness.CATS
        assert parser.parse_config(self.Cuteness.CATS) is self.Cuteness.CATS
        with pytest.raises(yuio.parse.ParsingError):
            parser.parse_config("CATS")
        assert parser.parse("dogs") is self.Cuteness.DOGS
        assert parser.parse(":3") is self.Cuteness.BLAHAJ
        with pytest.raises(yuio.parse.ParsingError):
            parser.parse("Unchi")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_by_name(self):
        parser = yuio.parse.Enum(self.Colors, by_name=True)
        assert parser.parse("RED") is self.Colors.RED
        assert parser.parse("RED") is self.Colors.RED
        assert parser.parse_config("RED") is self.Colors.RED
        assert parser.parse_config(self.Colors.RED) is self.Colors.RED
        assert parser.parse("green") is self.Colors.GREEN
        assert parser.parse("Blue") is self.Colors.BLUE
        with pytest.raises(yuio.parse.ParsingError):
            parser.parse("Color of a beautiful sunset")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

        assert parser.describe() == "{RED|GREEN|BLUE}"
        assert parser.describe_or_def() == "{RED|GREEN|BLUE}"
        assert parser.describe_value(self.Colors.RED) == "RED"

    def test_to_dash_case(self):
        class Colors(enum.Enum):
            RED = "RED"
            GREEN_FORE = "GREEN_FORE"
            GREEN_BACK = "GREEN_BACK"

        parser = yuio.parse.Enum(Colors, to_dash_case=True)
        assert parser.parse("green-fore") is Colors.GREEN_FORE

    def test_short(self):
        class Colors(enum.Enum):
            RED = "RED"
            GREEN_FORE = "GREEN_FORE"
            GREEN_BACK = "GREEN_BACK"

        parser = yuio.parse.Enum(Colors)
        assert parser.parse("R") is Colors.RED
        assert parser.parse("r") is Colors.RED
        with pytest.raises(
            yuio.parse.ParsingError,
            match=r"possible candidates are GREEN_FORE or GREEN_BACK",
        ):
            parser.parse("G")
        assert parser.parse("GREEN_F") is Colors.GREEN_FORE
        with pytest.raises(yuio.parse.ParsingError, match=r"did you mean RED?"):
            parser.parse_config("r")

    def test_from_type_hint(self):
        parser = yuio.parse.from_type_hint(self.Cuteness)
        assert isinstance(parser, yuio.parse.Enum)
        assert parser.parse(":3") is self.Cuteness.BLAHAJ

    def test_partial(self):
        parser = yuio.parse.Enum()
        with pytest.raises(TypeError, match=r"Enum requires an inner parser"):
            parser.parse("asd")  # type: ignore

        parser = yuio.parse.Enum(by_name=True)
        with pytest.raises(TypeError, match=r"Enum requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[self.Cuteness, yuio.parse.Enum()]
        )
        assert parser.parse(":3") is self.Cuteness.BLAHAJ

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Enum conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Enum()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Enum will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[self.Cuteness, yuio.parse.Enum(), yuio.parse.Enum()]
            )

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Enum with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[self.Colors, yuio.parse.Enum(self.Colors)]
            )


class TestDecimal:
    def test_basics(self):
        parser = yuio.parse.Decimal()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<decimal>"
        assert parser.describe_many() == "<decimal>"
        assert parser.describe_value(Decimal(10)) == "10"

    def test_json_schema(self):
        parser = yuio.parse.Decimal()
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref("#/$defs/Decimal", "Decimal")
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate(10)
        validator.validate(10.5)
        validator.validate("NaN")
        validator.validate("sNaN")
        validator.validate("-10e2")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        assert parser.to_json_value(Decimal(10)) == "10"

    def test_parse(self):
        parser = yuio.parse.Decimal()
        assert parser.parse("1.5") == Decimal("1.5")
        assert parser.parse("-10") == Decimal("-10")
        assert parser.parse_config(1.0) == Decimal("1.0")
        assert parser.parse_config("1.5") == Decimal("1.5")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("x")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse_config("x")
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Expected int, float, or str, got list"
        ):
            parser.parse_config([])

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(Decimal), yuio.parse.Decimal)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[Decimal, yuio.parse.Decimal()])
        assert parser.parse("1.5") == 1.5

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Decimal conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Decimal()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Decimal will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[Decimal, yuio.parse.Decimal(), yuio.parse.Decimal()]
            )


class TestFraction:
    def test_basics(self):
        parser = yuio.parse.Fraction()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<fraction>"
        assert parser.describe_many() == "<fraction>"
        assert parser.describe_value(Fraction(10)) == "10"

    def test_json_schema(self):
        parser = yuio.parse.Fraction()
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref("#/$defs/Fraction", "Fraction")
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate(10)
        validator.validate(10.5)
        validator.validate("NaN")
        validator.validate("1/2")
        validator.validate("-1/2")
        validator.validate("-10e2")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        assert parser.to_json_value(Fraction("2/3")) == "2/3"

    def test_parse(self):
        parser = yuio.parse.Fraction()
        assert parser.parse("1/3") == Fraction("1/3")
        assert parser.parse("-10") == Fraction("-10")
        assert parser.parse_config(1.0) == Fraction("1.0")
        assert parser.parse_config("1/3") == Fraction("1/3")
        assert parser.parse_config([2, 5]) == Fraction("2/5")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("x")
        with pytest.raises(yuio.parse.ParsingError, match=r"division by zero"):
            parser.parse("1/0")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse_config("x")
        with pytest.raises(
            yuio.parse.ParsingError,
            match=r"Expected int, float, str, or a tuple of two ints, got list",
        ):
            parser.parse_config([])
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Can't parse 1/0 as fraction"
        ):
            parser.parse_config([1, 0])
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Can't parse 1/nan as fraction"
        ):
            parser.parse_config([1, math.nan])
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Can't parse nan as fraction"
        ):
            parser.parse_config(math.nan)

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(Fraction), yuio.parse.Fraction)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[Fraction, yuio.parse.Fraction()]
        )
        assert parser.parse("1.5") == 1.5

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Fraction conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Fraction()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Fraction will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[Fraction, yuio.parse.Fraction(), yuio.parse.Fraction()]
            )


class TestJson:
    def test_basics(self):
        parser = yuio.parse.Json()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<json>"
        assert parser.describe_many() == "<json>"
        assert parser.describe_value(dict(a=0)) == "{'a': 0}"

    def test_unstructured(self):
        parser = yuio.parse.Json()
        assert parser.parse("[1, 2, 3]") == [1, 2, 3]
        assert parser.parse_config([1, 2, 3]) == [1, 2, 3]
        assert parser.parse_config("[1, 2, 3]") == "[1, 2, 3]"
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Can't parse 'x' as JsonValue"
        ):
            parser.parse("x")
        assert isinstance(
            yuio.parse.from_type_hint(yuio.json_schema.JsonValue), yuio.parse.Json
        )
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Any()
        )
        assert parser.to_json_value([1, dict(a=0)]) == [1, dict(a=0)]

    def test_structured(self):
        parser = yuio.parse.Json(yuio.parse.List(yuio.parse.Int()))
        assert parser.parse("[1, 2, 3]") == [1, 2, 3]
        assert parser.parse_config([1, 2, 3]) == [1, 2, 3]
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Can't parse 'x' as JsonValue"
        ):
            parser.parse("x")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected int, got float"):
            parser.parse("[1.5]")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected list, got str"):
            parser.parse_config("[1, 2, 3]")
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.Array(yuio.json_schema.Integer())
        assert parser.to_json_value([1, 2, 3]) == [1, 2, 3]
        with pytest.raises(TypeError):
            parser.to_json_value([1, dict(a=10)])

    def test_from_type_hint_annotated_unstructured(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[yuio.json_schema.JsonValue, yuio.parse.Json()]
        )
        assert parser.parse("1.5") == 1.5

    def test_from_type_hint_annotated_structured(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[list[int], yuio.parse.Json()])
        assert parser.parse("[1, 2, 3]") == [1, 2, 3]
        assert parser.parse_config([1, 2, 3]) == [1, 2, 3]
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Can't parse 'x' as JsonValue"
        ):
            parser.parse("x")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected list, got str"):
            parser.parse_config("[1, 2, 3]")

    def test_from_type_hint_annotated_shadowing(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[
                yuio.json_schema.JsonValue, yuio.parse.Json(), yuio.parse.Json()
            ]
        )
        assert parser.parse("[1, 2, 3]") == [1, 2, 3]
        assert parser.parse_config([1, 2, 3]) == [1, 2, 3]

    def test_from_type_hint_annotated_unstructured_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Json with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[
                    yuio.json_schema.JsonValue, yuio.parse.Json(yuio.parse.Int())
                ]
            )

    def test_from_type_hint_annotated_structured_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Json with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Json(yuio.parse.Int())]
            )


class TestDateTime:
    def test_basics(self):
        parser = yuio.parse.DateTime()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "YYYY-MM-DD[ HH:MM:SS]"
        assert parser.describe_or_def() == "YYYY-MM-DD[ HH:MM:SS]"
        assert parser.describe_many() == "YYYY-MM-DD[ HH:MM:SS]"
        assert (
            parser.describe_value(datetime.datetime(2025, 1, 1))
            == "2025-01-01 00:00:00"
        )

    def test_json_schema(self):
        parser = yuio.parse.DateTime()
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref("#/$defs/DateTime", "DateTime")
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate("2025-01-01 15:00")
        validator.validate("20250101T15:00")
        validator.validate("2025-W10-2")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("2025W10-2")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate(10)
        assert (
            parser.to_json_value(datetime.datetime(2025, 1, 1, 15))
            == "2025-01-01 15:00:00"
        )

    def test_parse(self):
        parser = yuio.parse.DateTime()

        assert parser.parse("2007-01-02T10:00:05.001") == datetime.datetime(
            2007, 1, 2, 10, 0, 5, 1000
        )
        assert parser.parse("2007-01-02 10:00:05.001") == datetime.datetime(
            2007, 1, 2, 10, 0, 5, 1000
        )
        assert parser.parse_config("2007-01-02T10:00:05.001") == datetime.datetime(
            2007, 1, 2, 10, 0, 5, 1000
        )
        assert parser.parse_config(
            datetime.datetime(2007, 1, 2, 10, 0, 5, 1000)
        ) == datetime.datetime(2007, 1, 2, 10, 0, 5, 1000)
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("2007 01 02")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_from_type_hint(self):
        assert isinstance(
            yuio.parse.from_type_hint(datetime.datetime), yuio.parse.DateTime
        )

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[datetime.datetime, yuio.parse.DateTime()]
        )
        assert parser.parse("2007-01-02 10:00:05.001") == datetime.datetime(
            2007, 1, 2, 10, 0, 5, 1000
        )

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with DateTime conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.DateTime()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with DateTime will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[
                    datetime.datetime, yuio.parse.DateTime(), yuio.parse.DateTime()
                ]
            )


class TestDate:
    def test_basics(self):
        parser = yuio.parse.Date()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "YYYY-MM-DD"
        assert parser.describe_or_def() == "YYYY-MM-DD"
        assert parser.describe_many() == "YYYY-MM-DD"
        assert parser.describe_value(datetime.date(2025, 1, 1)) == "2025-01-01"

    def test_json_schema(self):
        parser = yuio.parse.Date()
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref("#/$defs/Date", "Date")
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate("2025-01-01")
        validator.validate("20250101")
        validator.validate("2025-W10-2")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("2025W10-2")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate(10)
        assert parser.to_json_value(datetime.date(2025, 1, 1)) == "2025-01-01"

    def test_parse(self):
        parser = yuio.parse.Date()

        assert parser.parse("2007-01-02") == datetime.date(2007, 1, 2)
        assert parser.parse_config("2007-01-02") == datetime.date(2007, 1, 2)
        assert parser.parse_config(datetime.date(2007, 1, 2)) == datetime.date(
            2007, 1, 2
        )
        assert parser.parse_config(datetime.datetime(2007, 1, 2)) == datetime.date(
            2007, 1, 2
        )
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("2007 01 02")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(datetime.time(10, 1))

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(datetime.date), yuio.parse.Date)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[datetime.date, yuio.parse.Date()]
        )
        assert parser.parse("2007-01-02") == datetime.date(2007, 1, 2)

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Date conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Date()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Date will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[datetime.date, yuio.parse.Date(), yuio.parse.Date()]
            )


class TestTime:
    def test_basics(self):
        parser = yuio.parse.Time()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "HH:MM:SS"
        assert parser.describe_or_def() == "HH:MM:SS"
        assert parser.describe_many() == "HH:MM:SS"
        assert parser.describe_value(datetime.time()) == "00:00:00"

    def test_json_schema(self):
        parser = yuio.parse.Time()
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref("#/$defs/Time", "Time")
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate("15:00")
        validator.validate("15:00:00.123432Z")
        validator.validate("15:00:00-02")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("15:00:00-02Z")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate(10)
        assert parser.to_json_value(datetime.time(15, 00)) == "15:00:00"

    def test_parse(self):
        parser = yuio.parse.Time()

        assert parser.parse("10:05") == datetime.time(10, 5)
        assert parser.parse_config("10:05") == datetime.time(10, 5)
        assert parser.parse_config(datetime.time(10, 5)) == datetime.time(10, 5)
        assert parser.parse_config(
            datetime.datetime(2007, 1, 2, 12, 30, 5)
        ) == datetime.time(12, 30, 5)
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("10?05")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(datetime.date(1996, 1, 1))

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(datetime.time), yuio.parse.Time)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[datetime.time, yuio.parse.Time()]
        )
        assert parser.parse("10:00:05.001") == datetime.time(10, 0, 5, 1000)

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Time conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Time()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Time will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[datetime.time, yuio.parse.Time(), yuio.parse.Time()]
            )


class TestTimeDelta:
    def test_basics(self):
        parser = yuio.parse.TimeDelta()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "[+|-]HH:MM:SS"
        assert parser.describe_or_def() == "[+|-]HH:MM:SS"
        assert parser.describe_many() == "[+|-]HH:MM:SS"
        assert parser.describe_value(datetime.timedelta(hours=10)) == "10:00:00"

    def test_json_schema(self):
        parser = yuio.parse.TimeDelta()
        ctx = yuio.json_schema.JsonSchemaContext()
        res = parser.to_json_schema(ctx)
        assert res == yuio.json_schema.Ref("#/$defs/TimeDelta", "TimeDelta")
        schema: _t.Any = ctx.render(res)
        validator = jsonschema.Draft7Validator(schema)
        validator.validate("1day")
        validator.validate("-1 day")
        validator.validate("2 weeks 1 day 15:00:00")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("what?")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate("2 weeks 1 day,")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate(", 15:00:00")
        with pytest.raises(jsonschema.ValidationError):
            validator.validate(10)
        assert parser.to_json_value(datetime.timedelta(1, 25)) == "1 day, 0:00:25"

    def test_parse(self):
        parser = yuio.parse.TimeDelta()

        assert parser.parse("1w") == datetime.timedelta(weeks=1)
        assert parser.parse("1week") == datetime.timedelta(weeks=1)
        assert parser.parse("5weeks") == datetime.timedelta(weeks=5)
        assert parser.parse("2d") == datetime.timedelta(days=2)
        assert parser.parse("1day") == datetime.timedelta(days=1)
        assert parser.parse("3days") == datetime.timedelta(days=3)
        assert parser.parse("5s") == datetime.timedelta(seconds=5)
        assert parser.parse("1sec") == datetime.timedelta(seconds=1)
        assert parser.parse("2second") == datetime.timedelta(seconds=2)
        assert parser.parse("3secs") == datetime.timedelta(seconds=3)
        assert parser.parse("4seconds") == datetime.timedelta(seconds=4)
        assert parser.parse("5m") == datetime.timedelta(minutes=5)
        assert parser.parse("1min") == datetime.timedelta(minutes=1)
        assert parser.parse("2minute") == datetime.timedelta(minutes=2)
        assert parser.parse("3mins") == datetime.timedelta(minutes=3)
        assert parser.parse("4minutes") == datetime.timedelta(minutes=4)
        assert parser.parse("5h") == datetime.timedelta(hours=5)
        assert parser.parse("1h") == datetime.timedelta(hours=1)
        assert parser.parse("1hr") == datetime.timedelta(hours=1)
        assert parser.parse("2hour") == datetime.timedelta(hours=2)
        assert parser.parse("3hrs") == datetime.timedelta(hours=3)
        assert parser.parse("4hours") == datetime.timedelta(hours=4)
        assert parser.parse("5ms") == datetime.timedelta(milliseconds=5)
        assert parser.parse("5l") == datetime.timedelta(milliseconds=5)
        assert parser.parse("1milli") == datetime.timedelta(milliseconds=1)
        assert parser.parse("2millis") == datetime.timedelta(milliseconds=2)
        assert parser.parse("3millisecond") == datetime.timedelta(milliseconds=3)
        assert parser.parse("4milliseconds") == datetime.timedelta(milliseconds=4)
        assert parser.parse("5us") == datetime.timedelta(microseconds=5)
        assert parser.parse("5u") == datetime.timedelta(microseconds=5)
        assert parser.parse("1micro") == datetime.timedelta(microseconds=1)
        assert parser.parse("2micros") == datetime.timedelta(microseconds=2)
        assert parser.parse("3microsecond") == datetime.timedelta(microseconds=3)
        assert parser.parse("4microseconds") == datetime.timedelta(microseconds=4)

        assert parser.parse("1d5secs") == datetime.timedelta(1, 5)
        assert parser.parse("1d 5secs") == datetime.timedelta(1, 5)
        assert parser.parse("-1d5secs") == -datetime.timedelta(1, 5)
        assert parser.parse("- 1d 5secs") == -datetime.timedelta(1, 5)

        assert parser.parse("10:00:01") == datetime.timedelta(hours=10, seconds=1)
        assert parser.parse("-10:00:01") == datetime.timedelta(hours=-10, seconds=-1)

        assert parser.parse("1d 05:00") == datetime.timedelta(days=1, hours=5)
        assert parser.parse("+1d +05:00") == datetime.timedelta(days=1, hours=5)
        assert parser.parse("-1d 05:00") == datetime.timedelta(days=-1, hours=5)
        assert parser.parse("-1d +05:00") == datetime.timedelta(days=-1, hours=5)
        assert parser.parse("-1d -05:00") == datetime.timedelta(days=-1, hours=-5)
        assert parser.parse("-1d2h +05:00") == datetime.timedelta(days=-1, hours=-2 + 5)

        assert parser.parse("1d, +5:00") == datetime.timedelta(days=1, hours=5)

        with pytest.raises(yuio.parse.ParsingError, match=r"empty timedelta"):
            parser.parse("")

        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("-")

        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("00:00,")

        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse(",00:00")

        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("00")

        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("1 megasecond")

        assert parser.parse_config("10:00:01") == datetime.timedelta(
            hours=10, seconds=1
        )
        assert parser.parse_config(
            datetime.timedelta(hours=10, seconds=1)
        ) == datetime.timedelta(hours=10, seconds=1)
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str, got int: 10"):
            parser.parse_config(10)

    def test_from_type_hint(self):
        assert isinstance(
            yuio.parse.from_type_hint(datetime.timedelta), yuio.parse.TimeDelta
        )

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[datetime.timedelta, yuio.parse.TimeDelta()]
        )
        assert parser.parse("01:00") == datetime.timedelta(hours=1)

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with TimeDelta conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.TimeDelta()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with TimeDelta will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[
                    datetime.timedelta, yuio.parse.TimeDelta(), yuio.parse.TimeDelta()
                ]
            )


class TestPath:
    def test_basics(self):
        parser = yuio.parse.Path()
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<path>"
        assert parser.describe_many() == "<path>"
        assert parser.describe_value(pathlib.Path("/")) == os.path.sep

        parser = yuio.parse.Path(extensions=[".toml"])
        assert parser.describe_or_def() == "<*.toml>"
        assert parser.describe_or_def() == "<*.toml>"
        parser = yuio.parse.Path(extensions=[".toml", ".yaml"])
        assert parser.describe_or_def() == "{<*.toml>|<*.yaml>}"
        assert parser.describe_or_def() == "{<*.toml>|<*.yaml>}"

    def test_json_schema(self):
        parser = yuio.parse.Path()
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.String()
        )
        assert parser.to_json_value(pathlib.Path("path")) == "path"

    def test_parse(self):
        parser = yuio.parse.Path()
        assert parser.parse("/a/s/d") == pathlib.Path("/a/s/d").resolve().absolute()
        assert parser.parse("/a/s/d/..") == pathlib.Path("/a/s").resolve().absolute()
        assert parser.parse("a/s/d") == pathlib.Path("a/s/d").resolve().absolute()
        assert parser.parse("./a/s/./d") == pathlib.Path("a/s/d").resolve().absolute()
        assert (
            parser.parse("~/a")
            == pathlib.Path(os.path.expanduser("~/a")).resolve().absolute()
        )
        assert (
            parser.parse_config("/a/s/d") == pathlib.Path("/a/s/d").resolve().absolute()
        )
        assert (
            parser.parse_config("/a/s/d/..")
            == pathlib.Path("/a/s").resolve().absolute()
        )
        assert (
            parser.parse_config("a/s/d") == pathlib.Path("a/s/d").resolve().absolute()
        )
        assert (
            parser.parse_config("./a/s/./d")
            == pathlib.Path("a/s/d").resolve().absolute()
        )
        assert (
            parser.parse_config("~/a")
            == pathlib.Path(os.path.expanduser("~/a")).resolve().absolute()
        )
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected str"):
            parser.parse_config(10)

    def test_extensions(self):
        parser = yuio.parse.Path(extensions=[".cfg", ".txt"])
        assert (
            parser.parse("/a/s/d.cfg")
            == pathlib.Path("/a/s/d.cfg").resolve().absolute()
        )
        assert (
            parser.parse("/a/s/d.txt")
            == pathlib.Path("/a/s/d.txt").resolve().absolute()
        )
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should have extension \.cfg or \.txt"
        ):
            parser.parse("file.sql")

    def test_from_type_hint(self):
        assert isinstance(yuio.parse.from_type_hint(pathlib.Path), yuio.parse.Path)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[pathlib.Path, yuio.parse.Path()]
        )
        assert parser.parse("/a/s/d") == pathlib.Path("/a/s/d").resolve().absolute()

    def test_from_type_hint_annotated_extensions(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[pathlib.Path, yuio.parse.Path(extensions=[".x"])]
        )
        assert parser.parse("/a/s/d.x") == pathlib.Path("/a/s/d.x").resolve().absolute()
        with pytest.raises(yuio.parse.ParsingError, match=r"should have extension \.x"):
            parser.parse("file.y")

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Path conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Path()])

    def test_from_type_hint_annotated_shadowing(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Path will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[pathlib.Path, yuio.parse.Path(), yuio.parse.Path()]
            )


class TestExistingPath:
    def test_parse(self, tmpdir):
        tmpdir.join("file.cfg").write("hi!")

        parser = yuio.parse.ExistingPath()
        with pytest.raises(yuio.parse.ParsingError, match=r"doesn't exist"):
            parser.parse(tmpdir.join("file2.cfg").strpath)
        assert (
            str(parser.parse(tmpdir.join("file.cfg").strpath))
            == tmpdir.join("file.cfg").strpath
        )

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[pathlib.Path, yuio.parse.ExistingPath()]
        )
        assert isinstance(parser, yuio.parse.ExistingPath)


class TestNonExistentPath:
    def test_parse(self, tmpdir):
        tmpdir.join("file.cfg").write("hi!")

        parser = yuio.parse.NonExistentPath()
        with pytest.raises(yuio.parse.ParsingError, match=r"already exists"):
            parser.parse(tmpdir.join("file.cfg").strpath)
        assert (
            str(parser.parse(tmpdir.join("file2.cfg").strpath))
            == tmpdir.join("file2.cfg").strpath
        )

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[pathlib.Path, yuio.parse.NonExistentPath()]
        )
        assert isinstance(parser, yuio.parse.NonExistentPath)


class TestFile:
    def test_parse(self, tmpdir):
        tmpdir.join("file.cfg").write("hi!")

        parser = yuio.parse.File()
        assert (
            str(parser.parse(tmpdir.join("file.cfg").strpath))
            == tmpdir.join("file.cfg").strpath
        )
        with pytest.raises(yuio.parse.ParsingError, match=r"doesn't exist"):
            parser.parse(tmpdir.join("file.txt").strpath)
        with pytest.raises(yuio.parse.ParsingError, match=r"is not a file"):
            parser.parse(tmpdir.strpath)

        parser = yuio.parse.File(extensions=[".cfg", ".txt"])
        assert (
            str(parser.parse(tmpdir.join("file.cfg").strpath))
            == tmpdir.join("file.cfg").strpath
        )
        with pytest.raises(yuio.parse.ParsingError, match=r"doesn't exist"):
            parser.parse(tmpdir.join("file.txt").strpath)
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should have extension \.cfg or \.txt"
        ):
            parser.parse(tmpdir.join("file.sql").strpath)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[pathlib.Path, yuio.parse.File()]
        )
        assert isinstance(parser, yuio.parse.File)


class TestDir:
    def test_parse(self, tmpdir):
        tmpdir.join("file.cfg").write("hi!")

        parser = yuio.parse.Dir()
        assert str(parser.parse("~")) == os.path.expanduser("~")
        assert str(parser.parse(tmpdir.strpath)) == tmpdir.strpath
        with pytest.raises(yuio.parse.ParsingError, match=r"doesn't exist"):
            parser.parse(tmpdir.join("subdir").strpath)
        with pytest.raises(yuio.parse.ParsingError, match=r"is not a directory"):
            parser.parse(tmpdir.join("file.cfg").strpath)

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[pathlib.Path, yuio.parse.Dir()])
        assert isinstance(parser, yuio.parse.Dir)


class TestGitRepo:
    def test_parse(self, tmpdir):
        tmpdir.join("file.cfg").write("hi!")

        parser = yuio.parse.GitRepo()
        with pytest.raises(yuio.parse.ParsingError, match=r"is not a git repository"):
            parser.parse(tmpdir.strpath)
        with pytest.raises(yuio.parse.ParsingError, match=r"doesn't exist"):
            parser.parse(tmpdir.join("subdir").strpath)
        with pytest.raises(yuio.parse.ParsingError, match=r"is not a directory"):
            parser.parse(tmpdir.join("file.cfg").strpath)

        tmpdir.join(".git").mkdir()

        assert str(parser.parse(tmpdir.strpath)) == tmpdir.strpath

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[pathlib.Path, yuio.parse.GitRepo()]
        )
        assert isinstance(parser, yuio.parse.GitRepo)


class TestBound:
    def test_api(self):
        with pytest.raises(TypeError, match=r"lower and lower_inclusive"):
            yuio.parse.Bound(yuio.parse.Int(), lower=0, lower_inclusive=1)
        with pytest.raises(TypeError, match=r"upper and upper_inclusive"):
            yuio.parse.Bound(yuio.parse.Int(), upper=0, upper_inclusive=1)

    def test_empty_bounds(self):
        parser = yuio.parse.Bound(yuio.parse.Int())
        assert parser.parse("0") == 0
        assert parser.parse("-10") == -10
        assert parser.parse("10") == 10

    def test_gt(self):
        parser = yuio.parse.Gt(yuio.parse.Int(), 0)
        assert parser.parse("10") == 10
        with pytest.raises(yuio.parse.ParsingError, match=r"should be greater than 0"):
            parser.parse("-1")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be greater than 0"):
            parser.parse("0")

    def test_ge(self):
        parser = yuio.parse.Ge(yuio.parse.Int(), 0)
        assert parser.parse("10") == 10
        assert parser.parse("0") == 0
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should be greater than or equal to 0"
        ):
            parser.parse("-1")

    def test_lt(self):
        parser = yuio.parse.Lt(yuio.parse.Int(), 10)
        assert parser.parse("5") == 5
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 10"):
            parser.parse("10")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 10"):
            parser.parse("11")

    def test_le(self):
        parser = yuio.parse.Le(yuio.parse.Int(), 10)
        assert parser.parse("5") == 5
        assert parser.parse("10") == 10
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should be lesser than or equal to 10"
        ):
            parser.parse("11")

    def test_range_inclusive(self):
        parser = yuio.parse.Bound(
            yuio.parse.Int(), lower_inclusive=0, upper_inclusive=5
        )
        assert parser.parse("0") == 0
        assert parser.parse("2") == 2
        assert parser.parse("5") == 5
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should be greater than or equal to 0"
        ):
            parser.parse("-1")
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should be lesser than or equal to 5"
        ):
            parser.parse("6")

    def test_range_non_inclusive(self):
        parser = yuio.parse.Bound(yuio.parse.Int(), lower=0, upper=5)
        assert parser.parse("2") == 2
        with pytest.raises(yuio.parse.ParsingError, match=r"should be greater than 0"):
            parser.parse("0")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 5"):
            parser.parse("5")

    def test_partial(self):
        parser = yuio.parse.Bound()
        with pytest.raises(TypeError, match=r"Bound requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[int, yuio.parse.Bound(lower_inclusive=0, upper=10)]
        )
        assert parser.parse("0") == 0
        assert parser.parse("9") == 9
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 10"):
            parser.parse("10")

    def test_from_type_hint_annotated_combine(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[
                int, yuio.parse.Bound(lower_inclusive=0), yuio.parse.Bound(upper=10)
            ]
        )
        assert parser.parse("0") == 0
        assert parser.parse("9") == 9
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 10"):
            parser.parse("10")

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Bound with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Bound(yuio.parse.Int())]
            )

    def test_gt_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Gt(0)])
        assert parser.parse("10") == 10
        with pytest.raises(yuio.parse.ParsingError, match=r"should be greater than 0"):
            parser.parse("-1")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be greater than 0"):
            parser.parse("0")
        with pytest.raises(TypeError):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Gt(yuio.parse.Int(), 0, 0)]  # type: ignore
            )

    def test_ge_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Ge(0)])
        assert parser.parse("10") == 10
        assert parser.parse("0") == 0
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should be greater than or equal to 0"
        ):
            parser.parse("-1")
        with pytest.raises(TypeError):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Ge(yuio.parse.Int(), 0, 0)]  # type: ignore
            )

    def test_lt_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Lt(10)])
        assert parser.parse("5") == 5
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 10"):
            parser.parse("10")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be lesser than 10"):
            parser.parse("11")
        with pytest.raises(TypeError):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Lt(yuio.parse.Int(), 10, 0)]  # type: ignore
            )

    def test_le_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Le(10)])
        assert parser.parse("5") == 5
        assert parser.parse("10") == 10
        with pytest.raises(
            yuio.parse.ParsingError, match=r"should be lesser than or equal to 10"
        ):
            parser.parse("11")
        with pytest.raises(TypeError):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Le(yuio.parse.Int(), 10, 0)]  # type: ignore
            )


class TestOneOf:
    def test_parse(self):
        parser = yuio.parse.OneOf(yuio.parse.Str(), ["qux", "duo"])
        assert parser.parse("qux") == "qux"
        assert parser.parse("duo") == "duo"
        with pytest.raises(yuio.parse.ParsingError, match=r"should be 'qux' or 'duo'"):
            parser.parse("foo")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be 'qux' or 'duo'"):
            parser.parse("Qux")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be 'qux' or 'duo'"):
            parser.parse("Duo")

    def test_partial(self):
        parser = yuio.parse.OneOf([])
        with pytest.raises(TypeError, match=r"OneOf requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_one_of_combined_with_lower(self):
        parser = yuio.parse.OneOf(yuio.parse.Lower(yuio.parse.Str()), ["qux", "duo"])
        assert parser.parse("Qux") == "qux"
        assert parser.parse("Duo") == "duo"

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[str, yuio.parse.OneOf(["qux", "duo"])]
        )
        assert parser.parse("qux") == "qux"
        assert parser.parse("duo") == "duo"
        with pytest.raises(yuio.parse.ParsingError, match=r"should be 'qux' or 'duo'"):
            parser.parse("foo")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be 'qux' or 'duo'"):
            parser.parse("Qux")
        with pytest.raises(yuio.parse.ParsingError, match=r"should be 'qux' or 'duo'"):
            parser.parse("Duo")

    def test_from_type_hint_annotated_combined_with_lower(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[
                str,
                yuio.parse.Str(),
                yuio.parse.Lower(),
                yuio.parse.OneOf(["qux", "duo"]),
            ]
        )
        assert parser.parse("Qux") == "qux"
        assert parser.parse("Duo") == "duo"

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using OneOf with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.OneOf(yuio.parse.Int(), [1, 2, 3])]
            )


class TestMap:
    @dataclass
    class Wrapper:
        value: _t.Any

    def test_basics(self):
        parser = yuio.parse.Map(
            yuio.parse.Int(), TestMap.Wrapper, lambda wrapped: wrapped.value
        )
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<int>"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(TestMap.Wrapper(10)) == "10"

    def test_secret(self):
        parser = yuio.parse.Map(yuio.parse.Secret(yuio.parse.Str()), lambda x: x)
        assert parser.is_secret()

    def test_basics_collection(self):
        parser = yuio.parse.Map(
            yuio.parse.List(yuio.parse.Int()),
            TestMap.Wrapper,
            lambda wrapped: wrapped.value,
        )
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.options() is None
        assert parser.parse_many(["1", "2", "3"]) == TestMap.Wrapper([1, 2, 3])
        assert parser.describe() == "<int>[ <int>[ ...]]"
        assert parser.describe_or_def() == "<int>[ <int>[ ...]]"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(TestMap.Wrapper([])) == ""

    def test_same_type(self):
        parser = yuio.parse.Map(yuio.parse.Int(), lambda x: x * 2)
        assert parser.describe_value(20) == "20"

        parser = yuio.parse.Map(yuio.parse.Int(), lambda x: x * 2, lambda x: x // 2)
        assert parser.describe_value(20) == "10"

    def test_json_schema(self):
        parser = yuio.parse.Map(yuio.parse.Bool(), lambda x: not x, lambda x: not x)
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Boolean()
        )
        assert not parser.to_json_value(True)

    def test_options(self):
        class E(enum.Enum):
            pass

        parser = yuio.parse.Map(yuio.parse.Enum(E), lambda x: x)
        assert parser.options() == []

        class E(enum.Enum):
            FOO = "foo"

        parser = yuio.parse.Map(yuio.parse.Enum(E), lambda x: x)
        assert parser.options() == [
            yuio.widget.Option(
                value=E.FOO,
                display_text="foo",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
        ]

    def test_parse(self):
        parser = yuio.parse.Map(yuio.parse.Int(), lambda x: x * 2)
        assert parser.parse("2") == 4
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("foo")

    def test_partial(self):
        parser = yuio.parse.Map(lambda x: x)
        with pytest.raises(TypeError, match=r"Map requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[int, yuio.parse.Map[int, int](lambda x: x * 2)]
        )
        assert parser.parse("2") == 4
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("foo")

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Map with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Map(yuio.parse.Int(), lambda x: x)]
            )


class TestApply:
    def test_basics(self):
        parser = yuio.parse.Apply(yuio.parse.Int(), lambda x: None)
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<int>"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(10) == "10"

    def test_secret(self):
        parser = yuio.parse.Apply(yuio.parse.Secret(yuio.parse.Str()), lambda x: None)
        assert parser.is_secret()

    def test_basics_collection(self):
        parser = yuio.parse.Apply(yuio.parse.List(yuio.parse.Int()), lambda x: None)
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.parse_many(["1", "2", "3"]) == [1, 2, 3]
        assert parser.describe() == "<int>[ <int>[ ...]]"
        assert parser.describe_or_def() == "<int>[ <int>[ ...]]"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value([]) == ""

    def test_json_schema(self):
        parser = yuio.parse.Apply(yuio.parse.Int(), lambda x: None)
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Integer()
        )
        assert parser.to_json_value(10) == 10

    def test_options(self):
        class E(enum.Enum):
            pass

        parser = yuio.parse.Apply(yuio.parse.Enum(E), lambda x: None)
        assert parser.options() == []

        class E(enum.Enum):
            FOO = "foo"

        parser = yuio.parse.Apply(yuio.parse.Enum(E), lambda x: None)
        assert parser.options() == [
            yuio.widget.Option(
                value=E.FOO,
                display_text="foo",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
        ]

    def test_parse(self):
        value = None

        def fn(x):
            nonlocal value
            value = x
            return x * 2

        parser = yuio.parse.Apply(yuio.parse.Int(), fn)
        assert parser.parse("2") == 2
        assert value == 2

        value = None

        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("foo")
        assert value is None

    def test_partial(self):
        parser = yuio.parse.Apply(print)
        with pytest.raises(TypeError, match=r"Apply requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        value = None

        def fn(x):
            nonlocal value
            value = x
            return x * 2

        parser = yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Apply(fn)])
        assert parser.parse("2") == 2
        assert value == 2

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Apply with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Apply(yuio.parse.Int(), lambda x: None)]
            )


class TestSimpleCollections:
    @pytest.fixture(
        params=[
            (yuio.parse.List, list),
            (yuio.parse.Set, set),
            (yuio.parse.FrozenSet, frozenset),
        ]
    )
    def test_params(self, request):
        return request.param

    def test_parse(self, test_params):
        parser_cls, ctor = test_params

        parser = parser_cls(yuio.parse.Int())
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.options() is None
        assert not parser.is_secret()
        assert parser.parse("1 2 3") == ctor([1, 2, 3])
        assert parser.parse_config([1, 2, 3]) == ctor([1, 2, 3])
        assert parser.parse_many(["1", "2", "3"]) == ctor([1, 2, 3])
        assert parser.describe() == "<int>[ <int>[ ...]]"
        assert parser.describe_or_def() == "<int>[ <int>[ ...]]"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(ctor([1, 2, 3])) == "1 2 3"

        with pytest.raises(yuio.parse.ParsingError, match=r"Expected .*?, got int"):
            parser.parse_config(123)

    def test_secret(self, test_params):
        parser_cls, _ = test_params

        parser = parser_cls(yuio.parse.Secret(yuio.parse.Int()))
        assert parser.is_secret()

    def test_json_schema(self, test_params):
        parser_cls, ctor = test_params
        parser = parser_cls(yuio.parse.Int())
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.Array(
            yuio.json_schema.Integer(), unique_items=issubclass(ctor, (set, frozenset))
        )
        assert parser.to_json_value(ctor([1, 2, 3])) == [1, 2, 3]

    def test_delim(self, test_params):
        parser_cls, ctor = test_params

        parser = parser_cls(yuio.parse.Int(), delimiter=",")
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.parse("1, 2, 3") == ctor([1, 2, 3])
        assert parser.parse_config([1, 2, 3]) == ctor([1, 2, 3])
        assert parser.parse_many(["1", "2", "3"]) == ctor([1, 2, 3])
        assert parser.describe() == "<int>[,<int>[,...]]"
        assert parser.describe_or_def() == "<int>[,<int>[,...]]"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(ctor([1, 2, 3])) == "1,2,3"

    def test_partial(self, test_params):
        parser_cls, _ = test_params

        parser = parser_cls()
        with pytest.raises(TypeError, match=r"requires an inner parser"):
            parser.parse("asd")

    def test_from_type_hint_annotated(self, test_params):
        parser_cls, ctor = test_params
        parser = yuio.parse.from_type_hint(_t.Annotated[ctor[int], parser_cls()])
        assert parser.parse("1 2 3") == ctor([1, 2, 3])
        parser = yuio.parse.from_type_hint(
            _t.Annotated[ctor[int], parser_cls(delimiter=",")]
        )
        assert parser.parse("1, 2, 3") == ctor([1, 2, 3])

    def test_from_type_hint_annotated_shadows(self, test_params):
        parser_cls, ctor = test_params
        with pytest.raises(
            TypeError,
            match=r"annotating a type with .*? will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[ctor[int], parser_cls(), parser_cls()]
            )

    def test_from_type_hint_annotated_wrong_type(self, test_params):
        parser_cls, _ = test_params
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with .*? conflicts with "
                "default parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, parser_cls()])

    def test_from_type_hint_annotated_non_partial(self, test_params):
        parser_cls, ctor = test_params
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using .*? with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[ctor[int], parser_cls(yuio.parse.Int())]
            )


class TestDict:
    def test_json_schema(self):
        parser = yuio.parse.Dict(yuio.parse.Str(), yuio.parse.Int())
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.Dict(
            yuio.json_schema.String(), yuio.json_schema.Integer()
        )
        assert parser.to_json_value({"x": 1}) == {"x": 1}

        parser = yuio.parse.Dict(yuio.parse.Int(), yuio.parse.Str())
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.Dict(
            yuio.json_schema.Integer(), yuio.json_schema.String()
        )
        assert parser.to_json_value({1: "x"}) == [[1, "x"]]

    def test_secret(self):
        parser = yuio.parse.Dict(
            yuio.parse.Secret(yuio.parse.Str()),
            yuio.parse.Str(),
        )
        assert parser.is_secret()

        parser = yuio.parse.Dict(
            yuio.parse.Str(),
            yuio.parse.Secret(yuio.parse.Str()),
        )
        assert parser.is_secret()

    def test_parse(self):
        parser = yuio.parse.Dict(yuio.parse.Int(), yuio.parse.Str())
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.options() is None
        assert not parser.is_secret()
        assert parser.parse("1:a 2:b") == {1: "a", 2: "b"}
        assert parser.parse_config({1: "a", 2: "b"}) == {1: "a", 2: "b"}
        assert parser.parse_config([(1, "a"), (2, "b")]) == {1: "a", 2: "b"}
        assert parser.parse_many(["1:a", "2:b"]) == {1: "a", 2: "b"}
        assert parser.describe() == "<int>:<str>[ <int>:<str>[ ...]]"
        assert parser.describe_or_def() == "<int>:<str>[ <int>:<str>[ ...]]"
        assert parser.describe_many() == "<int>:<str>"
        assert parser.describe_value({1: "a", 2: "b"}) == "1:a 2:b"

        with pytest.raises(
            yuio.parse.ParsingError, match=r"Expected dict or list, got int"
        ):
            parser.parse_config(123)
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected int, got str"):
            parser.parse_config({"x": "y"})

    def test_delim(self):
        parser = yuio.parse.Dict(
            yuio.parse.Int(), yuio.parse.Str(), delimiter=",", pair_delimiter="-"
        )
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.parse("1-a, 2-b") == {1: "a", 2: "b"}
        assert parser.parse_many(["1-a", "2-b"]) == {1: "a", 2: "b"}
        assert parser.describe() == "<int>-<str>[,<int>-<str>[,...]]"
        assert parser.describe_or_def() == "<int>-<str>[,<int>-<str>[,...]]"
        assert parser.describe_many() == "<int>-<str>"
        assert parser.describe_value({1: "a", 2: "b"}) == "1-a,2-b"

    def test_partial(self):
        parser = yuio.parse.Dict()
        with pytest.raises(TypeError, match=r"requires an inner parser"):
            parser.parse("asd")  # type: ignore
        parser = yuio.parse.Dict(yuio.parse.Int())  # type: ignore
        with pytest.raises(TypeError, match=r"requires an inner parser"):
            parser.parse("asd")

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[dict[int, str], yuio.parse.Dict()]
        )
        assert parser.parse("1:a 2:b") == {1: "a", 2: "b"}
        parser = yuio.parse.from_type_hint(
            _t.Annotated[
                dict[int, str], yuio.parse.Dict(delimiter=",", pair_delimiter="-")
            ]
        )
        assert parser.parse("1-a,2-b") == {1: "a", 2: "b"}

    def test_from_type_hint_annotated_shadows(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Dict will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[dict[int, str], yuio.parse.Dict(), yuio.parse.Dict()]
            )

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Dict conflicts with "
                "default parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Dict()])

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Dict with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[
                    dict[str, int],
                    yuio.parse.Dict(yuio.parse.Str(), yuio.parse.Int()),
                ]
            )


class TestTuple:
    def test_json_schema(self):
        parser = yuio.parse.Tuple(yuio.parse.Str(), yuio.parse.Int(), yuio.parse.Bool())
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.Tuple(
            [
                yuio.json_schema.String(),
                yuio.json_schema.Integer(),
                yuio.json_schema.Boolean(),
            ]
        )
        assert parser.to_json_value(("x", -5, True)) == ["x", -5, True]

    def test_basics(self):
        parser = yuio.parse.Tuple(yuio.parse.Int())
        assert parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        assert parser.parse_many(["1"]) == (1,)
        assert parser.describe() == "<int>"
        assert parser.describe_or_def() == "<int>"
        assert parser.describe_many() == ("<int>",)
        assert parser.describe_value((10,)) == "10"

        parser = yuio.parse.Tuple(yuio.parse.Int(), yuio.parse.Str())
        assert parser.supports_parse_many()
        assert parser.get_nargs() == 2
        assert parser.options() is None
        assert parser.parse_many(["1", "x"]) == (1, "x")
        assert parser.describe() == "<int> <str>"
        assert parser.describe_or_def() == "<int> <str>"
        assert parser.describe_many() == ("<int>", "<str>")
        assert parser.describe_value((10, "x")) == "10 x"

        parser = yuio.parse.Tuple(yuio.parse.Int(), yuio.parse.Str(), yuio.parse.Bool())
        assert parser.supports_parse_many()
        assert parser.get_nargs() == 3
        assert parser.options() is None
        assert parser.parse_many(["1", "x", "no"]) == (1, "x", False)
        assert parser.describe() == "<int> <str> {yes|no}"
        assert parser.describe_or_def() == "<int> <str> {yes|no}"
        assert parser.describe_many() == ("<int>", "<str>", "{yes|no}")
        assert parser.describe_value((10, "x", False)) == "10 x no"

    def test_secret(self):
        parser = yuio.parse.Tuple(yuio.parse.Secret(yuio.parse.Str()))
        assert parser.is_secret()
        parser = yuio.parse.Tuple(
            yuio.parse.Int(), yuio.parse.Str(), yuio.parse.Secret(yuio.parse.Str())
        )
        assert parser.is_secret()

    def test_parse(self):
        parser = yuio.parse.Tuple(yuio.parse.Int())
        assert parser.parse("1") == (1,)
        assert parser.parse_config([1]) == (1,)
        assert parser.parse_config([1.0]) == (1,)
        assert parser.parse_many(["1"]) == (1,)
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("1 2 3")
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse"):
            parser.parse("foo")
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Expected list or tuple, got int"
        ):
            parser.parse_config(5)
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected int, got str"):
            parser.parse_config(["5"])
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected 1 element, got 0"):
            parser.parse("")
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected 1 element, got 2"):
            parser.parse_config([1, 2])
        with pytest.raises(yuio.parse.ParsingError, match=r"Expected 1 element, got 2"):
            parser.parse_many(["1", "2"])

        parser = yuio.parse.Tuple(yuio.parse.Str(), yuio.parse.Str())
        assert parser.parse("1 2") == ("1", "2")
        assert parser.parse("x y z") == ("x", "y z")
        assert parser.parse_config(["x", "y"]) == ("x", "y")
        assert parser.parse_many(["x", "y"]) == ("x", "y")
        with pytest.raises(
            yuio.parse.ParsingError, match=r"Expected 2 elements, got 1"
        ):
            parser.parse("x")

    def test_delim(self):
        parser = yuio.parse.Tuple(yuio.parse.Str(), yuio.parse.Str(), delimiter=",")
        assert parser.parse("1, 2") == ("1", " 2")
        assert parser.parse("x, y, z") == ("x", " y, z")
        assert parser.parse_config(["x", "y"]) == ("x", "y")
        assert parser.parse_many(["x", "y"]) == ("x", "y")

    def test_partial(self):
        parser = yuio.parse.Tuple()
        with pytest.raises(TypeError, match=r"requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[tuple[int, str], yuio.parse.Tuple()]
        )
        assert parser.parse("1 a") == (1, "a")
        parser = yuio.parse.from_type_hint(
            _t.Annotated[tuple[int, str], yuio.parse.Tuple(delimiter=",")]
        )
        assert parser.parse("1,a") == (1, "a")

    def test_from_type_hint_annotated_shadows(self):
        with pytest.raises(
            TypeError,
            match=r"annotating a type with Tuple will override all previous annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[tuple[int, str], yuio.parse.Tuple(), yuio.parse.Tuple()]
            )

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                r"annotating list(\[int\])? with Tuple conflicts with "
                r"default parser for this type, which is List\."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[list[int], yuio.parse.Tuple()])

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using Tuple with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[
                    tuple[str, int],
                    yuio.parse.Tuple(yuio.parse.Str(), yuio.parse.Int()),
                ]
            )


class TestOptional:
    def test_basics(self):
        parser = yuio.parse.Optional(yuio.parse.Int())
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<int>"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value(10) == "10"
        assert parser.describe_value(None) == "<none>"

    def test_secret(self):
        parser = yuio.parse.Optional(yuio.parse.Secret(yuio.parse.Str()))
        assert parser.is_secret()

    def test_json_schema(self):
        parser = yuio.parse.Optional(yuio.parse.Int())
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.OneOf(
            [yuio.json_schema.Integer(), yuio.json_schema.Null()]
        )
        assert parser.to_json_value(None) is None
        assert parser.to_json_value(10) == 10

    def test_basics_collection(self):
        parser = yuio.parse.Optional(yuio.parse.List(yuio.parse.Int()))
        assert parser.supports_parse_many()
        assert parser.get_nargs() == "*"
        assert parser.parse_many(["1", "2", "3"]) == [1, 2, 3]
        assert parser.describe() == "<int>[ <int>[ ...]]"
        assert parser.describe_or_def() == "<int>[ <int>[ ...]]"
        assert parser.describe_many() == "<int>"
        assert parser.describe_value([]) == ""
        assert parser.describe_value(None) == "<none>"

    def test_options(self):
        class E(enum.Enum):
            pass

        parser = yuio.parse.Optional(yuio.parse.Enum(E))
        assert parser.options() == []

        class E(enum.Enum):
            FOO = "foo"

        parser = yuio.parse.Optional(yuio.parse.Enum(E))
        assert parser.options() == [
            yuio.widget.Option(
                value=E.FOO,
                display_text="foo",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
        ]

    def test_parse(self):
        parser = yuio.parse.Optional(yuio.parse.Int())
        assert parser.parse("1") == 1
        assert parser.parse_config(1) == 1
        assert parser.parse_config(None) is None
        assert parser.supports_parse_many() is False

    def test_partial(self):
        parser = yuio.parse.Optional()
        with pytest.raises(TypeError, match=r"requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[list[int], yuio.parse.Optional()]
        )
        assert parser.parse_config(None) is None
        assert parser.parse_config([1, 2, 3]) == [1, 2, 3]

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using .*? with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[int, yuio.parse.Optional(yuio.parse.Int())]
            )


class TestUnion:
    def test_basics(self):
        parser = yuio.parse.Union(yuio.parse.Int(), yuio.parse.Str())
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "{<int>|<str>}"
        assert parser.describe_or_def() == "{<int>|<str>}"
        assert parser.describe_many() == "{<int>|<str>}"
        assert parser.describe_value(10) == "10"

    def test_secret(self):
        parser = yuio.parse.Union(
            yuio.parse.Secret(yuio.parse.Str()),
        )
        assert parser.is_secret()
        parser = yuio.parse.Union(
            yuio.parse.Secret(yuio.parse.Str()),
            yuio.parse.Int(),
        )
        assert parser.is_secret()
        parser = yuio.parse.Union(
            yuio.parse.Int(),
            yuio.parse.Secret(yuio.parse.Str()),
        )
        assert parser.is_secret()

    def test_json_schema(self):
        parser = yuio.parse.Union(
            yuio.parse.Map(yuio.parse.Int(), lambda x: x * 2, lambda x: x // 2),
            yuio.parse.Date(),
            yuio.parse.Str(),
        )
        assert parser.to_json_schema(
            yuio.json_schema.JsonSchemaContext()
        ) == yuio.json_schema.OneOf(
            [
                yuio.json_schema.Integer(),
                yuio.json_schema.Ref("#/$defs/Date", "Date"),
                yuio.json_schema.String(),
            ]
        )
        assert parser.to_json_value("10") == "10"
        assert parser.to_json_value(datetime.date(2024, 1, 1)) == "2024-01-01"
        assert parser.to_json_value(10) == 5

    def test_options(self):
        class E(enum.Enum):
            pass

        parser = yuio.parse.Union(yuio.parse.Enum(E))
        assert parser.options() is None

        class E(enum.Enum):
            FOO = "foo"

        parser = yuio.parse.Union(yuio.parse.Enum(E))
        assert parser.options() == [
            yuio.widget.Option(
                value=E.FOO,
                display_text="foo",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
        ]

        class E2(enum.Enum):
            BAR = "bar"

        parser = yuio.parse.Union(yuio.parse.Enum(E), yuio.parse.Enum(E2))
        assert parser.options() == [
            yuio.widget.Option(
                value=E.FOO,
                display_text="foo",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
            yuio.widget.Option(
                value=E2.BAR,
                display_text="bar",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
        ]

    def test_parse(self):
        parser = yuio.parse.Union(
            yuio.parse.Int(), yuio.parse.OneOf(yuio.parse.Str(), ["foo", "bar"])
        )

        assert parser.parse("10") == 10
        assert parser.parse("foo") == "foo"
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse 'baz'"):
            parser.parse("baz")

        assert parser.parse_config(10) == 10
        assert parser.parse_config("foo") == "foo"
        with pytest.raises(yuio.parse.ParsingError, match=r"Can't parse 'baz'"):
            parser.parse_config("baz")

    def test_partial(self):
        parser = yuio.parse.Union()
        with pytest.raises(TypeError, match=r"requires an inner parser"):
            parser.parse("asd")  # type: ignore

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(_t.Annotated[int | str, yuio.parse.Union()])
        assert parser.parse("10") == 10
        assert parser.parse("foo") == "foo"

    def test_from_type_hint_annotated_wrong_type(self):
        with pytest.raises(
            TypeError,
            match=(
                "annotating int with Union conflicts with default "
                "parser for this type, which is Int."
            ),
        ):
            yuio.parse.from_type_hint(_t.Annotated[int, yuio.parse.Union()])

    def test_from_type_hint_annotated_non_partial(self):
        with pytest.raises(
            TypeError,
            match=r"don't provide inner parser when using .*? with type annotations",
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[
                    int | str,
                    yuio.parse.Union(yuio.parse.Int(), yuio.parse.Str()),
                ]
            )


class TestWithMeta:
    def test_basics(self):
        parser = yuio.parse.WithMeta(yuio.parse.Int(), desc="desc")
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert not parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() == "desc"
        assert parser.describe_or_def() == "desc"
        assert parser.describe_many() == "desc"
        assert parser.describe_value(10) == "10"

    def test_secret(self):
        parser = yuio.parse.WithMeta(yuio.parse.Secret(yuio.parse.Str()), desc="...")
        assert parser.is_secret()

    def test_json_schema(self):
        parser = yuio.parse.WithMeta(yuio.parse.Int(), desc="desc")
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.Integer()
        )
        assert parser.to_json_value(10) == 10

    def test_options(self):
        class E(enum.Enum):
            pass

        parser = yuio.parse.WithMeta(yuio.parse.Enum(E))
        assert parser.options() == []

        class E(enum.Enum):
            FOO = "foo"

        parser = yuio.parse.WithMeta(yuio.parse.Enum(E))
        assert parser.options() == [
            yuio.widget.Option(
                value=E.FOO,
                display_text="foo",
                display_text_prefix="",
                display_text_suffix="",
                comment=None,
                color_tag="none",
            ),
        ]

    def test_parse(self):
        parser = yuio.parse.WithMeta(yuio.parse.Int(), desc="desc")
        assert parser.parse("1") == 1
        assert parser.parse_config(1) == 1

    def test_from_type_hint_annotated(self):
        parser = yuio.parse.from_type_hint(
            _t.Annotated[int, yuio.parse.WithMeta(desc="desc")]
        )
        assert parser.describe() == "desc"

    def test_from_type_hint_annotated_not_partial(self):
        with pytest.raises(
            TypeError,
            match=(
                "don't provide inner parser when using WithMeta with type annotations"
            ),
        ):
            yuio.parse.from_type_hint(
                _t.Annotated[str, yuio.parse.WithMeta(yuio.parse.Int(), desc="desc")]
            )


class TestSecret:
    def test_basics(self):
        parser = yuio.parse.Secret(yuio.parse.Str())
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == 1
        assert parser.options() is None
        assert parser.is_secret()
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["1", "2", "3"])
        assert parser.describe() is None
        assert parser.describe_or_def() == "<str>"
        assert parser.describe_many() == "<str>"
        assert parser.describe_value(10) == "***"

    def test_from_type_hint(self):
        parser = yuio.parse.from_type_hint(yuio.parse.SecretString)
        assert isinstance(parser, yuio.parse.Secret)
        assert isinstance(parser._inner_raw, yuio.parse.Str)
        with pytest.raises(TypeError, match=r"SecretValue requires type arguments"):
            yuio.parse.from_type_hint(yuio.parse.SecretValue)

    def test_options(self):
        class E(enum.Enum):
            pass

        parser = yuio.parse.Secret(yuio.parse.Enum(E))
        assert parser.options() is None

        class E(enum.Enum):
            FOO = "foo"

        parser = yuio.parse.Secret(yuio.parse.Enum(E))
        assert parser.options() is None

    def test_parse(self):
        parser = yuio.parse.Secret(yuio.parse.Int())

        assert parser.parse("10") == yuio.secret.SecretValue(10)
        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
        ):
            parser.parse("xxx")

    def test_parse_config(self):
        parser = yuio.parse.Secret(yuio.parse.Int())

        assert parser.parse_config(10) == yuio.secret.SecretValue(10)
        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
            match=r"^Expected int, got str$",
        ):
            parser.parse_config("xxx")

    def test_parse_many(self):
        parser = yuio.parse.Secret(yuio.parse.List(yuio.parse.Int()))

        assert parser.parse("10 11 12") == yuio.secret.SecretValue([10, 11, 12])
        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
            match=r"^Can't parse value as int$",
        ):
            parser.parse("xxx")

        assert parser.parse_many(["10", "11", "12"]) == yuio.secret.SecretValue(
            [10, 11, 12]
        )
        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
            match=r"^Can't parse value as int$",
        ):
            parser.parse_many(["10", "xxx", "12"])

    def test_list_of_secrets(self):
        parser = yuio.parse.List(yuio.parse.Secret(yuio.parse.Int()))
        assert parser.is_secret()
        assert parser.parse("10 11 12") == [
            yuio.secret.SecretValue(10),
            yuio.secret.SecretValue(11),
            yuio.secret.SecretValue(12),
        ]

        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
            match=r"^Can't parse value as int$",
        ):
            parser.parse("xxx")

        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
            match=r"^Can't parse value as int$",
        ):
            parser.parse_many(["10", "xxx", "12"])

        with pytest.raises(
            yuio.parse.ParsingError,
            check=lambda e: "xxx" not in str(e) and e.raw is None,
            match=r"^In \$\[1\]:\n  Expected int, got str$",
        ):
            parser.parse_config([10, "xxx", 12])



import yuio.app
import yuio.color
import yuio.io
import yuio.term
import yuio.theme

class Theme(yuio.theme.DefaultTheme):
    def __init__(self, term: yuio.term.Term):
        super().__init__(term)

        if not term.terminal_theme:
            return

        background = term.terminal_theme.background
        magenta = term.terminal_theme.magenta

        match term.terminal_theme.lightness:
            case yuio.term.Lightness.LIGHT:
                # Slightly darker than background.
                dark_magenta = magenta.match_luminosity(background.darken(0.2))
            case yuio.term.Lightness.DARK:
                # Slightly lighter than background.
                dark_magenta = magenta.match_luminosity(background.lighten(0.2))
            case _:
                # As dark as background.
                dark_magenta = magenta.match_luminosity(background)

        self._set_color_if_not_overridden(
            "dark_magenta_bg", yuio.color.Color(back=dark_magenta)
        )

@yuio.app.app
def main():
    yuio.io.info("This is a <c dark_magenta_bg>dark magenta background!</c>")

main.theme = Theme

if __name__ == "__main__":
    main.run()



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Utility functions and types.

"""

from __future__ import annotations

import re as _re
import textwrap as _textwrap

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

__all__ = [
    "UserString",
    "dedent",
    "to_dash_case",
]

_UNPRINTABLE = "".join([chr(i) for i in range(32)]) + "\x7f"
_UNPRINTABLE_TRANS = str.maketrans(_UNPRINTABLE, " " * len(_UNPRINTABLE))
_UNPRINTABLE_RE = r"[" + _re.escape(_UNPRINTABLE) + "]"
_UNPRINTABLE_RE_WITHOUT_NL = r"[" + _re.escape(_UNPRINTABLE.replace("\n", "")) + "]"

_TO_DASH_CASE_RE = _re.compile(
    r"""
      # We will add a dash (bear with me here):
      [_\s]                             # 1. instead of underscore or space,
      | (                               # 2. OR in the following case:
        (?<!^)                          #   - not at the beginning of the string,
        (                               #   - AND EITHER:
            (?<=[A-Z])(?=[A-Z][a-z])    #     - before case gets lower (`XMLTag` -> `XML-Tag`),
          | (?<=[a-zA-Z])(?![a-zA-Z_])  #     - between a letter and a non-letter (`HTTP20` -> `HTTP-20`),
          | (?<![A-Z_])(?=[A-Z])        #     - between non-uppercase and uppercase letter (`TagXML` -> `Tag-XML`),
        )                               #   - AND ALSO:
        (?!$)                           #     - not at the end of the string.
      )
    """,
    _re.VERBOSE | _re.MULTILINE,
)


def to_dash_case(msg: str, /) -> str:
    """
    Convert ``CamelCase`` or ``snake_case`` identifier to a ``dash-case`` one.

    This function assumes ASCII input, and will not work correctly
    with non-ASCII characters.

    :param msg:
        identifier to convert.
    :returns:
        identifier in ``dash-case``.
    :example:
        ::

            >>> to_dash_case("SomeClass")
            'some-class'
            >>> to_dash_case("HTTP20XMLUberParser")
            'http-20-xml-uber-parser'

    """

    return _TO_DASH_CASE_RE.sub("-", msg).lower()


def dedent(msg: str, /):
    """
    Remove leading indentation from a message and normalize trailing newlines.

    This function is intended to be used with triple-quote string literals,
    such as docstrings. It will remove common indentation from second
    and subsequent lines, then it will strip any leading and trailing whitespaces
    and add a new line at the end.

    :param msg:
        message to dedent.
    :returns:
        normalized message.
    :example:
        ::

            >>> def foo():
            ...     \"""Documentation for function ``foo``.
            ...
            ...     Leading indent is stripped.
            ...     \"""
            ...
            ...     ...

            >>> dedent(foo.__doc__)
            'Documentation for function ``foo``.\\n\\nLeading indent is stripped.\\n'

    """

    if not msg:
        return msg

    first, *rest = msg.splitlines(keepends=True)
    return (first.rstrip() + "\n" + _textwrap.dedent("".join(rest))).strip() + "\n"


_COMMENT_RE = _re.compile(r"^\s*#:(.*)\r?\n?$")
_RST_ROLE_RE = _re.compile(
    r"(?::[\w+.:-]+:|__?)?`((?:[^`\n\\]|\\.)+)`(?::[\w+.:-]+:|__?)?", _re.DOTALL
)
_RST_ROLE_TITLE_RE = _re.compile(
    r"^((?:[^`\n\\]|\\.)*) <(?:[^`\n\\]|\\.)*>$", _re.DOTALL
)
_ESC_RE = _re.compile(r"\\(.)", _re.DOTALL)


def _rst_esc_repl(match: _re.Match[str]):
    symbol = match.group(1)
    if symbol in "\n\r\t\v\b":
        return " "
    return symbol


def _rst_repl(match: _re.Match[str]):
    full: str = match.group(0)
    text: str = match.group(1)
    if full.startswith(":") or full.endswith(":"):
        if title_match := _RST_ROLE_TITLE_RE.match(text):
            text = title_match.group(1)
        elif text.startswith("~"):
            text = text.rsplit(".", maxsplit=1)[-1]
    text = _ESC_RE.sub(_rst_esc_repl, text)
    n_backticks = 0
    cur_n_backticks = 0
    for ch in text:
        if ch == "`":
            cur_n_backticks += 1
        else:
            n_backticks = max(cur_n_backticks, n_backticks)
            cur_n_backticks = 0
    n_backticks = max(cur_n_backticks, n_backticks)
    if not n_backticks:
        return f"`{text}`"
    else:
        bt = "`" * (n_backticks + 1)
        return f"{bt} {text} {bt}"


def _process_docstring(msg: str, /, only_first_paragraph: bool = True):
    value = dedent(msg).removesuffix("\n")

    if only_first_paragraph and (index := value.find("\n\n")) != -1:
        value = value[:index]

    return _RST_ROLE_RE.sub(_rst_repl, value)


def _find_docs(obj: _t.Any, /) -> dict[str, str]:
    """
    Find documentation for fields of a class.

    Inspects source code of a class and finds docstrings and doc comments (``#:``)
    for variables in its body. Doesn't inspect ``__init__``, doesn't return documentation
    for class methods. Returns first paragraph from each docstring, formatted for use
    in CLI help messages.

    """

    # Based on code from Sphinx, two clause BSD license.
    # See https://github.com/sphinx-doc/sphinx/blob/master/LICENSE.rst.

    import ast
    import inspect
    import itertools

    if (qualname := getattr(obj, "__qualname__", None)) is None:
        # Not a known object.
        return {}

    if "<locals>" in qualname:
        # This will not work as expected!
        return {}

    try:
        sourcelines, _ = inspect.getsourcelines(obj)
    except TypeError:
        return {}

    docs: dict[str, str] = {}

    node = ast.parse(_textwrap.dedent("".join(sourcelines)))
    assert isinstance(node, ast.Module)
    assert len(node.body) == 1
    cdef = node.body[0]

    if isinstance(cdef, ast.ClassDef):
        fields: list[tuple[int, str]] = []
        last_field: str | None = None
        for stmt in cdef.body:
            if (
                last_field
                and isinstance(stmt, ast.Expr)
                and isinstance(stmt.value, ast.Constant)
                and isinstance(stmt.value.value, str)
            ):
                docs[last_field] = _process_docstring(stmt.value.value)
            last_field = None
            if isinstance(stmt, ast.AnnAssign):
                target = stmt.target
            elif isinstance(stmt, ast.Assign) and len(stmt.targets) == 1:
                target = stmt.targets[0]
            else:
                continue
            if isinstance(target, ast.Name) and not target.id.startswith("_"):
                fields.append((stmt.lineno, target.id))
                last_field = target.id
    elif isinstance(cdef, ast.FunctionDef):
        fields = [
            (field.lineno, field.arg)
            for field in itertools.chain(cdef.args.args, cdef.args.kwonlyargs)
        ]
    else:  # pragma: no cover
        return {}

    for pos, name in fields:
        comment_lines: list[str] = []
        for before_line in sourcelines[pos - 2 :: -1]:
            if match := _COMMENT_RE.match(before_line):
                comment_lines.append(match.group(1))
            else:
                break

        if comment_lines:
            docs[name] = _process_docstring("\n".join(reversed(comment_lines)))

    return docs


if TYPE_CHECKING:

    class _FormatMapMapping(_t.Protocol):
        def __getitem__(self, key: str, /) -> _t.Any: ...

    class _TranslateTable(_t.Protocol):
        def __getitem__(self, key: int, /) -> str | int | None: ...


class UserString(str):
    """
    Base class for user string.

    This class is similar to :class:`collections.UserString`, but actually derived
    from string, with customizable wrapping semantics, and returns custom string
    instances from all string methods (:class:`collections.UserString` doesn't
    wrap strings returned from :meth:`str.split` and similar).

    .. tip::

        When deriving from this class, add ``__slots__`` to avoid making a string
        with a ``__dict__`` property.

    .. seealso::

        See implementation of :class:`yuio.string.Link` for an example of handling user
        strings with internal state.

    """

    __slots__ = ()

    def _wrap(self, data: str) -> _t.Self:
        """
        Wrap raw string that resulted from an operation on this instance into another
        instance of :class:`UserString`.

        Override this method if you need to preserve some internal state during
        operations.

        By default, this simply creates an instance of ``self.__class__`` with the
        given string.

        """

        return self.__class__(data)

    def __add__(self, value: str, /) -> _t.Self:
        return self._wrap(super().__add__(value))

    def __format__(self, format_spec: str, /) -> _t.Self:
        return self._wrap(super().__format__(format_spec))

    def __getitem__(self, key: _t.SupportsIndex | slice, /) -> _t.Self:
        return self._wrap(super().__getitem__(key))

    def __mod__(self, value: _t.Any, /) -> _t.Self:
        return self._wrap(super().__mod__(value))

    def __mul__(self, value: _t.SupportsIndex, /) -> _t.Self:
        return self._wrap(super().__mul__(value))

    def __rmul__(self, value: _t.SupportsIndex, /) -> _t.Self:
        return self._wrap(super().__rmul__(value))

    def capitalize(self) -> _t.Self:
        return self._wrap(super().capitalize())

    def casefold(self) -> _t.Self:
        return self._wrap(super().casefold())

    def center(self, width: _t.SupportsIndex, fillchar: str = " ", /) -> _t.Self:
        return self._wrap(super().center(width))

    def expandtabs(self, tabsize: _t.SupportsIndex = 8) -> _t.Self:
        return self._wrap(super().expandtabs(tabsize))

    def format_map(self, mapping: _FormatMapMapping, /) -> _t.Self:
        return self._wrap(super().format_map(mapping))

    def format(self, *args: object, **kwargs: object) -> _t.Self:
        return self._wrap(super().format(*args, **kwargs))

    def join(self, iterable: _t.Iterable[str], /) -> _t.Self:
        return self._wrap(super().join(iterable))

    def ljust(self, width: _t.SupportsIndex, fillchar: str = " ", /) -> _t.Self:
        return self._wrap(super().ljust(width, fillchar))

    def lower(self) -> _t.Self:
        return self._wrap(super().lower())

    def lstrip(self, chars: str | None = None, /) -> _t.Self:
        return self._wrap(super().lstrip(chars))

    def partition(self, sep: str, /) -> tuple[_t.Self, _t.Self, _t.Self]:
        l, c, r = super().partition(sep)
        return self._wrap(l), self._wrap(c), self._wrap(r)

    def removeprefix(self, prefix: str, /) -> _t.Self:
        return self._wrap(super().removeprefix(prefix))

    def removesuffix(self, suffix: str, /) -> _t.Self:
        return self._wrap(super().removesuffix(suffix))

    def replace(self, old: str, new: str, count: _t.SupportsIndex = -1, /) -> _t.Self:
        return self._wrap(super().replace(old, new, count))

    def rjust(self, width: _t.SupportsIndex, fillchar: str = " ", /) -> _t.Self:
        return self._wrap(super().rjust(width, fillchar))

    def rpartition(self, sep: str, /) -> tuple[_t.Self, _t.Self, _t.Self]:
        l, c, r = super().rpartition(sep)
        return self._wrap(l), self._wrap(c), self._wrap(r)

    def rsplit(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, sep: str | None = None, maxsplit: _t.SupportsIndex = -1
    ) -> list[_t.Self]:
        return [self._wrap(part) for part in super().rsplit(sep, maxsplit)]

    def rstrip(self, chars: str | None = None, /) -> _t.Self:
        return self._wrap(super().rstrip(chars))

    def split(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, sep: str | None = None, maxsplit: _t.SupportsIndex = -1
    ) -> list[_t.Self]:
        return [self._wrap(part) for part in super().split(sep, maxsplit)]

    def splitlines(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, keepends: bool = False
    ) -> list[_t.Self]:
        return [self._wrap(part) for part in super().splitlines(keepends)]

    def strip(self, chars: str | None = None, /) -> _t.Self:
        return self._wrap(super().strip(chars))

    def swapcase(self) -> _t.Self:
        return self._wrap(super().swapcase())

    def title(self) -> _t.Self:
        return self._wrap(super().title())

    def translate(self, table: _TranslateTable, /) -> _t.Self:
        return self._wrap(super().translate(table))

    def upper(self) -> _t.Self:
        return self._wrap(super().upper())

    def zfill(self, width: _t.SupportsIndex, /) -> _t.Self:
        return self._wrap(super().zfill(width))


T = _t.TypeVar("T", covariant=True)



import yuio.string
from yuio.color import Color
from yuio.string import NO_WRAP_END, NO_WRAP_START

import pytest


@pytest.fixture(autouse=True)
def setup_repr_hl(theme):
    theme.set_color("hl/kwd:repr", Color.FORE_RED)
    theme.set_color("hl/str:repr", Color.FORE_GREEN)
    theme.set_color("hl/str/esc:repr", Color.FORE_BLUE)
    theme.set_color("hl/punct:repr", Color.FORE_CYAN)
    theme.set_color("hl/comment:repr", Color.FORE_MAGENTA)
    theme.set_color("hl/lit:repr", Color.FORE_YELLOW)
    theme.set_color("hl/type:repr", Color.FORE_BLACK)
    theme.set_color("hl/more:repr", Color.FORE_WHITE)


def _join_consecutive_strings(l):
    r = []
    for x in l:
        if isinstance(x, str) and r and isinstance(r[-1], str):
            r[-1] += x
        else:
            r.append(x)
    return r


@pytest.mark.parametrize(
    ("text", "kwargs", "expect"),
    [
        pytest.param(
            t"foo bar",
            {},
            [Color.NONE, "foo bar"],
            id="simple",
        ),
        pytest.param(
            t"foo {'interpolation'} bar",
            {},
            [Color.NONE, "foo ", "interpolation", " bar"],
            id="interpolation",
        ),
        pytest.param(
            t"foo {'абв'!s} bar",
            {},
            [Color.NONE, "foo ", "абв", " bar"],
            id="str",
        ),
        pytest.param(
            t"foo {'абв'!r} bar",
            {},
            [Color.NONE, "foo ", "'абв'", " bar"],
            id="repr",
        ),
        pytest.param(
            t"foo {'абв'!a} bar",
            {},
            [
                Color.NONE,
                "foo ",
                "'\\u0430\\u0431\\u0432'",
                " bar",
            ],
            id="ascii",
        ),
        pytest.param(
            t"foo {(1, 2)!r:#} bar",
            {},
            [
                Color.NONE,
                "foo ",
                Color.FORE_CYAN,
                "(",
                Color.FORE_YELLOW,
                "1",
                Color.FORE_CYAN,
                ", ",
                Color.FORE_YELLOW,
                "2",
                Color.FORE_CYAN,
                ")",
                Color.NONE,
                " bar",
            ],
            id="colorized_repr",
        ),
        pytest.param(
            t"foo {(1, 2):#} bar",
            {},
            [
                Color.NONE,
                "foo ",
                Color.FORE_CYAN,
                "(",
                Color.FORE_YELLOW,
                "1",
                Color.FORE_CYAN,
                ", ",
                Color.FORE_YELLOW,
                "2",
                Color.FORE_CYAN,
                ")",
                Color.NONE,
                " bar",
            ],
            id="colorized_repr_fallback",
        ),
        pytest.param(
            t"foo {(1, 2)!r:+} bar",
            {},
            [
                Color.NONE,
                "foo ",
                "(\n  1,\n  2\n)",
                " bar",
            ],
            id="multiline_repr",
        ),
        pytest.param(
            t"foo {(1, 2)!r:#+} bar",
            {},
            [
                Color.NONE,
                "foo ",
                Color.FORE_CYAN,
                "(",
                Color.NONE,
                "\n  ",
                Color.FORE_YELLOW,
                "1",
                Color.FORE_CYAN,
                ",",
                Color.NONE,
                "\n  ",
                Color.FORE_YELLOW,
                "2",
                Color.NONE,
                "\n",
                Color.FORE_CYAN,
                ")",
                Color.NONE,
                " bar",
            ],
            id="colorized_multiline_repr",
        ),
        pytest.param(
            t"foo {10:+05} bar",
            {},
            [
                Color.NONE,
                "foo ",
                "+0010",
                " bar",
            ],
            id="overridden_format",
        ),
        pytest.param(
            t"<c bold magenta>foo {(1, 2)!r:#}</c> bar",
            {},
            [
                Color.FORE_MAGENTA | Color.STYLE_BOLD,
                "foo ",
                Color.FORE_CYAN | Color.STYLE_BOLD,
                "(",
                Color.FORE_YELLOW | Color.STYLE_BOLD,
                "1",
                Color.FORE_CYAN | Color.STYLE_BOLD,
                ", ",
                Color.FORE_YELLOW | Color.STYLE_BOLD,
                "2",
                Color.FORE_CYAN | Color.STYLE_BOLD,
                ")",
                Color.NONE,
                " bar",
            ],
            id="colorized_repr_with_background",
        ),
        pytest.param(
            t"foo {'param':10} bar",
            {},
            [Color.NONE, "foo ", "param     ", " bar"],
            id="width",
        ),
        pytest.param(
            t"foo {'param':~<10} bar",
            {},
            [Color.NONE, "foo ", "param~~~~~", " bar"],
            id="fill",
        ),
        pytest.param(
            t"foo {'param':<10} bar",
            {},
            [Color.NONE, "foo ", "param     ", " bar"],
            id="flush_left",
        ),
        pytest.param(
            t"foo {'param':>10} bar",
            {},
            [Color.NONE, "foo ", "     param", " bar"],
            id="flush_right",
        ),
        pytest.param(
            t"foo {'param':^10} bar",
            {},
            [Color.NONE, "foo ", "  param   ", " bar"],
            id="center",
        ),
        pytest.param(
            t"foo {'param':✨<10} bar",
            {},
            [
                Color.NONE,
                "foo ",
                "param✨✨ ",
                " bar",
            ],
            id="flush_left_wide",
        ),
        pytest.param(
            t"foo {'param':✨>10} bar",
            {},
            [
                Color.NONE,
                "foo ",
                "✨✨ param",
                " bar",
            ],
            id="flush_right_wide",
        ),
        pytest.param(
            t"foo {'param':✨^10} bar",
            {},
            [
                Color.NONE,
                "foo ",
                "✨param✨ ",
                " bar",
            ],
            id="center_wide",
        ),
        pytest.param(
            t"foo `{10}` bar",
            {},
            [
                Color.NONE,
                "foo ",
                NO_WRAP_START,
                Color.FORE_MAGENTA,
                "10",
                NO_WRAP_END,
                Color.NONE,
                " bar",
            ],
            id="code",
        ),
        pytest.param(
            t"foo `...{10}...` bar",
            {},
            [
                Color.NONE,
                "foo ",
                NO_WRAP_START,
                Color.FORE_MAGENTA,
                "...10...",
                NO_WRAP_END,
                Color.NONE,
                " bar",
            ],
            id="code_with_symbols",
        ),
        pytest.param(
            t"`{10}` bar",
            {},
            [NO_WRAP_START, Color.FORE_MAGENTA, "10", NO_WRAP_END, Color.NONE, " bar"],
            id="code_start",
        ),
        pytest.param(
            t"foo `{10}`",
            {},
            [Color.NONE, "foo ", NO_WRAP_START, Color.FORE_MAGENTA, "10", NO_WRAP_END],
            id="code_end",
        ),
        pytest.param(
            t"foo `{10} {11}` bar",
            {},
            [
                Color.NONE,
                "foo ",
                NO_WRAP_START,
                Color.FORE_MAGENTA,
                "10 11",
                NO_WRAP_END,
                Color.NONE,
                " bar",
            ],
            id="code_multiple_interpolations",
        ),
        pytest.param(
            t"foo ` {10} ` bar",
            {},
            [
                Color.NONE,
                "foo ",
                NO_WRAP_START,
                Color.FORE_MAGENTA,
                "10",
                NO_WRAP_END,
                Color.NONE,
                " bar",
            ],
            id="code_trim_spaces",
        ),
        pytest.param(
            t"foo `  {10}  ` bar",
            {},
            [
                Color.NONE,
                "foo ",
                NO_WRAP_START,
                Color.FORE_MAGENTA,
                " 10 ",
                NO_WRAP_END,
                Color.NONE,
                " bar",
            ],
            id="code_trim_spaces_2",
        ),
        pytest.param(
            t"foo ```{10}``` bar",
            {},
            [
                Color.NONE,
                "foo ",
                NO_WRAP_START,
                Color.FORE_MAGENTA,
                "10",
                NO_WRAP_END,
                Color.NONE,
                " bar",
            ],
            id="code_long_fence",
        ),
    ],
)
def test_colorize(text, kwargs, expect, ctx):
    formatted = yuio.string.colorize(text, ctx=ctx, **kwargs)
    assert _join_consecutive_strings(formatted._parts) == _join_consecutive_strings(
        expect
    )



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Yuio library. See documentation at https://yuio.readthedocs.io/.

"""

from __future__ import annotations

import enum as _enum
import logging as _logging
import os as _os
import sys as _sys
import warnings

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

if TYPE_CHECKING:
    import yuio.string

try:
    from yuio._version import *  # noqa: F403
except ImportError:
    raise ImportError(
        "yuio._version not found. if you are developing locally, "
        "run `pip install -e .` to generate it"
    )

__all__ = [
    "DISABLED",
    "GROUP",
    "MISSING",
    "POSITIONAL",
    "Disabled",
    "Group",
    "Missing",
    "Positional",
    "PrettyException",
    "YuioWarning",
    "enable_internal_logging",
]


class _Placeholders(_enum.Enum):
    DISABLED = "<disabled>"
    MISSING = "<missing>"
    POSITIONAL = "<positional>"
    GROUP = "<group>"

    def __bool__(self) -> _t.Literal[False]:
        return False  # pragma: no cover

    def __repr__(self):
        return f"yuio.{self.name}"  # pragma: no cover

    def __str__(self) -> str:
        return self.value  # pragma: no cover


Disabled: _t.TypeAlias = _t.Literal[_Placeholders.DISABLED]
"""
Type of the :data:`DISABLED` placeholder.

"""

DISABLED: Disabled = _Placeholders.DISABLED
"""
Indicates that some functionality is disabled.

"""


Missing: _t.TypeAlias = _t.Literal[_Placeholders.MISSING]
"""
Type of the :data:`MISSING` placeholder.

"""

MISSING: Missing = _Placeholders.MISSING
"""
Indicates that some value is missing.

"""


Positional: _t.TypeAlias = _t.Literal[_Placeholders.POSITIONAL]
"""
Type of the :data:`POSITIONAL` placeholder.

"""

POSITIONAL: Positional = _Placeholders.POSITIONAL
"""
Used with :func:`yuio.app.field` to enable positional arguments.

"""


Group: _t.TypeAlias = _t.Literal[_Placeholders.GROUP]
"""
Type of the :data:`GROUP` placeholder.

"""

GROUP: Group = _Placeholders.GROUP
"""
Used with :func:`yuio.app.field` to omit arguments from CLI usage.

"""


class YuioWarning(RuntimeWarning):
    """
    Base class for all runtime warnings.

    """


class PrettyException(Exception):
    """PrettyException(msg: typing.LiteralString, /, *args: typing.Any)
    PrettyException(msg: str, /)

    Base class for pretty-printable exceptions.

    :param msg:
        message to format. Can be a literal string or any other colorable object.

        If it's given as a literal string, additional arguments for ``%``-formatting
        may be given. Otherwise, giving additional arguments will cause
        a :class:`TypeError`.
    :param args:
        arguments for ``%``-formatting the message.
    :example:
        .. code-block:: python

        class MyError(yuio.PrettyException):
            pass


        try:
            ...
            raise MyError("A formatted <c b>error message</c>")
            ...
        except MyError as e:
            yuio.io.error_with_tb(e)

    """

    @_t.overload
    def __init__(self, msg: _t.LiteralString, /, *args): ...
    @_t.overload
    def __init__(self, msg: yuio.string.ToColorable | None = None, /): ...
    def __init__(self, *args):
        self.args = args

    def __rich_repr__(self) -> yuio.string.RichReprResult:
        yield from ((None, arg) for arg in self.args)

    def __str__(self) -> str:
        return str(self.to_colorable())

    def __colorized_str__(
        self, ctx: yuio.string.ReprContext
    ) -> yuio.string.ColorizedString:
        return ctx.str(self.to_colorable())

    def to_colorable(self) -> yuio.string.Colorable:
        """
        Return a colorable object with this exception's message.

        """

        if not self.args:
            return ""

        import yuio.string

        return yuio.string._to_colorable(self.args[0], self.args[1:])


_logger = _logging.getLogger("yuio.internal")
_logger.setLevel(_logging.DEBUG)
_logger.propagate = False

__stderr_handler = _logging.StreamHandler(_sys.__stderr__)
__stderr_handler.setLevel(_logging.CRITICAL)
_logger.addHandler(__stderr_handler)


def enable_internal_logging(
    path: str | None = None,
    level: str | int | None = None,
    propagate=None,
    add_handler: bool = False,
):  # pragma: no cover
    """
    Enable Yuio's internal logging.

    This function enables :func:`logging.captureWarnings`, and enables printing
    of :class:`YuioWarning` messages, and sets up logging channels ``yuio.internal``
    and ``py.warning``.

    :param path:
        if given, adds handlers that output internal log messages to the given file.
    :param level:
        configures logging level for file handler. Default is ``DEBUG``.
    :param propagate:
        if given, enables or disables log message propagation from ``yuio.internal``
        and ``py.warning`` to the root logger.
    :param add_handler:
        if :data:`True`, adds yuio handler to the logging. This is useful if you wish
        to see yuio log before main setup.

    """

    warn_logger = _logging.getLogger("py.warnings")

    if path:
        if level is None:
            level = _os.environ.get("YUIO_DEBUG", "").strip().upper() or "DEBUG"
        if level in ["1", "Y", "YES", "TRUE"]:
            level = "DEBUG"
        file_handler = _logging.FileHandler(path, delay=True)
        file_handler.setFormatter(
            _logging.Formatter("%(filename)s:%(lineno)d: %(levelname)s: %(message)s")
        )
        file_handler.setLevel(level)
        _logger.addHandler(file_handler)
        warn_logger.addHandler(file_handler)

    _logging.captureWarnings(True)
    warnings.simplefilter("default", category=YuioWarning)

    if propagate is not None:
        warn_logger.propagate = propagate
        _logger.propagate = propagate

    if add_handler:
        import yuio.io

        if not any(
            isinstance(handler, yuio.io.Handler) for handler in _logger.handlers
        ):
            _logger.addHandler(yuio.io.Handler())
        if not any(
            isinstance(handler, yuio.io.Handler) for handler in warn_logger.handlers
        ):
            warn_logger.addHandler(yuio.io.Handler())


_debug = "YUIO_DEBUG" in _os.environ or "YUIO_DEBUG_FILE" in _os.environ
if _debug:  # pragma: no cover
    enable_internal_logging(path=_os.environ.get("YUIO_DEBUG_FILE"), add_handler=True)
elif hasattr(_sys, "ps1"):
    enable_internal_logging(add_handler=True)
else:
    warnings.simplefilter("ignore", category=YuioWarning, append=True)
